# Base classes for various types of derived datasets - these are basically
# wrapper classes around data objects that modify the data in some way
# on-the-fly. These work by exposting the common data API described in
# http://docs.glueviz.org/en/stable/developer_guide/data.html and transparently
# applying changes.

from glue.core.hub import HubListener
from glue.core.data import BaseCartesianData
from glue.core.message import NumericalDataChangedMessage
from glue.core.subset import SliceSubsetState
from glue.core.component_id import ComponentID

from astropy.wcs.wcsapi import SlicedLowLevelWCS


class DerivedData(BaseCartesianData):
    """
    Base class for all derived data classes.
    """


class IndexedData(BaseCartesianData, HubListener):
    """
    A dataset where some dimensions have been removed via indexing.

    The indices can be modified after the data has been created using the
    ``indices`` property.

    Parameters
    ----------
    original_data : `~glue.core.data.BaseCartesianData`
        The original data to reduce the dimension of.
    indices : `tuple` of `int` or `None`
        The indices to apply to the data, or `None` if a dimension should be
        preserved. This tuple should have as many elements as there are
        dimensions in ``original_data``.
    """

    def __init__(self, original_data, indices):

        super(IndexedData, self).__init__()

        if len(indices) != original_data.ndim:
            raise ValueError("The 'indices' tuple should have length {0}"
                             .format(original_data.ndim))

        if hasattr(original_data, 'coords'):
            if original_data.coords is None:
                self._coords = None
            else:
                slices = [slice(None) if idx is None else idx for idx in indices]
                self._coords = SlicedLowLevelWCS(original_data.coords, slices)

        self._original_data = original_data
        self._cid_to_original_cid = {}
        self._original_cid_to_cid = {}
        self.indices = indices

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, value):

        if len(value) != self._original_data.ndim:
            raise ValueError("The 'indices' tuple should have length {0}"
                             .format(self._original_data.ndim))

        # For now we require the indices to be in the same position, i.e. we
        # don't allow changes in dimensionality of the derived dataset.
        if hasattr(self, '_indices'):
            changed = False
            for idim in range(self._original_data.ndim):
                before, after = self._indices[idim], value[idim]
                if type(before) is not type(after):
                    raise TypeError("Can't change where the ``None`` values are in indices")
                elif before != after:
                    changed = True
        else:
            changed = False

        self._indices = value

        # Compute a subset state that represents the indexing - this is used
        # for compute_statistic and compute_histogram
        slices = [slice(x) if x is None else x for x in self._indices]
        self._indices_subset_state = SliceSubsetState(self._original_data, slices)

        # Construct a list of original pixel component IDs
        self._original_pixel_cids = []
        for idim in range(self._original_data.ndim):
            if self._indices[idim] is None:
                self._original_pixel_cids.append(self._original_data.pixel_component_ids[idim])

        # Construct a list of original world component IDs
        self._original_world_cids = []
        if len(self._original_data.world_component_ids) > 0:
            idim_new = 0
            for idim in range(self._original_data.ndim):
                if self._indices[idim] is None:
                    self._cid_to_original_cid[self.world_component_ids[idim_new]] = self._original_data.world_component_ids[idim]
                    idim_new += 1

        # Tell glue that the data has changed
        if changed and self.hub is not None:
            msg = NumericalDataChangedMessage(self)
            self.hub.broadcast(msg)

    @property
    def label(self):
        slice = '[' + ','.join([':' if x is None else str(x) for x in self._indices]) + ']'
        return self._original_data.label + slice

    @property
    def shape(self):
        shape = []
        for idim in range(self._original_data.ndim):
            if self.indices[idim] is None:
                shape.append(self._original_data.shape[idim])
        return tuple(shape)

    @property
    def main_components(self):
        main = []
        for cid in self._original_data.main_components:
            if cid not in self._original_cid_to_cid:
                cid_new = ComponentID(label=cid.label, parent=self)
                self._original_cid_to_cid[cid] = cid_new
                self._cid_to_original_cid[cid_new] = cid
            main.append(self._original_cid_to_cid[cid])
        return main

    def get_kind(self, cid):
        cid = self._translate_cid(cid)
        return self._original_data.get_kind(cid)

    def _to_original_view(self, view):
        if view is None:
            view = [slice(None)] * self.ndim
        original_view = list(self.indices)
        idim_reduced = 0
        for idim in range(self._original_data.ndim):
            if original_view[idim] is None:
                if view is None:
                    original_view[idim] = slice(None)
                else:
                    original_view[idim] = view[idim_reduced]
                idim_reduced += 1
        return tuple(original_view)

    def register_to_hub(self, hub):
        hub.subscribe(self, NumericalDataChangedMessage,
                      handler=self._check_for_original_changes)
        super(IndexedData, self).register_to_hub(hub)

    def _check_for_original_changes(self, message):
        # If the parent's values are changed, we should assume the values
        # of the current dataset have changed too
        if message.data is self._original_data:
            msg = NumericalDataChangedMessage(self)
            self.hub.broadcast(msg)

    def _translate_cid(self, cid):
        if cid in self.pixel_component_ids:
            cid = self._original_pixel_cids[cid.axis]
        elif cid in self._cid_to_original_cid:
            cid = self._cid_to_original_cid[cid]
        return cid

    def get_data(self, cid, view=None):
        cid = self._translate_cid(cid)
        original_view = self._to_original_view(view)
        return self._original_data.get_data(cid, view=original_view)

    def get_mask(self, subset_state, view=None):
        original_view = self._to_original_view(view)
        return self._original_data.get_mask(subset_state, view=original_view)

    def compute_fixed_resolution_buffer(self, *args, **kwargs):
        if 'target_cid' in kwargs:
            kwargs['target_cid'] = self._translate_cid(kwargs['target_cid'])
        from glue.core.fixed_resolution_buffer import compute_fixed_resolution_buffer
        return compute_fixed_resolution_buffer(self, *args, **kwargs)

    def compute_statistic(self, statistic, cid, **kwargs):
        cid = self._translate_cid(cid)
        kwargs['view'] = self._to_original_view(kwargs.get('view'))
        return self._original_data.compute_statistic(statistic, cid, **kwargs)

    def compute_histogram(self, *args, **kwargs):
        if kwargs.get('subset_state') is None:
            kwargs['subset_state'] = self._indices_subset_state
        else:
            kwargs['subset_state'] &= self._indices_subset_state
        return self._original_data.compute_histogram(*args, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from functools import wraps

import pytest

try:
    import pytest_mpl  # noqa
except ImportError:
    HAS_PYTEST_MPL = False
else:
    HAS_PYTEST_MPL = True


def visual_test(*args, **kwargs):
    """
    A decorator that defines a visual test.

    This automatically decorates tests with mpl_image_compare with common
    options used by all figure tests in glue-core.
    """

    tolerance = kwargs.pop("tolerance", 0)
    style = kwargs.pop("style", {})
    savefig_kwargs = kwargs.pop("savefig_kwargs", {})
    savefig_kwargs["metadata"] = {"Software": None}

    def decorator(test_function):
        @pytest.mark.mpl_image_compare(
            tolerance=tolerance, style=style, savefig_kwargs=savefig_kwargs, **kwargs
        )
        @pytest.mark.skipif(
            not HAS_PYTEST_MPL, reason="pytest-mpl is required for the figure tests"
        )
        @wraps(test_function)
        def test_wrapper(*args, **kwargs):
            return test_function(*args, **kwargs)

        return test_wrapper

    # If the decorator was used without any arguments, the only positional
    # argument will be the test to decorate so we do the following:
    if len(args) == 1:
        return decorator(*args)

    return decorator



from glue.core.visual import VisualAttributes
from glue.utils.matplotlib import MATPLOTLIB_GE_36

if MATPLOTLIB_GE_36:
    from matplotlib import colormaps
else:
    from matplotlib.cm import get_cmap

import pytest


def test_VA_preferred_cmap():
    # Not a real CMAP array - errors
    with pytest.raises(TypeError, match="`preferred_cmap` must be a string or an instance of "
                       "a matplotlib.colors.Colormap"):
        VisualAttributes(preferred_cmap=1)

    # Not a valid string / known key [mpl 3.6+] for a CMAP - errors
    with pytest.raises(ValueError, match="not_a_cmap is not a valid colormap name."):
        VisualAttributes(preferred_cmap="not_a_cmap")

    viridis_cmap = colormaps["viridis"] if MATPLOTLIB_GE_36 else get_cmap("viridis")

    # get_cmap cmap name
    va = VisualAttributes(preferred_cmap="viridis")
    assert va.preferred_cmap == viridis_cmap
    # formal cmap name
    va = VisualAttributes(preferred_cmap="Viridis")
    assert va.preferred_cmap == viridis_cmap

    # Valid Colormap
    va = VisualAttributes(preferred_cmap=viridis_cmap)
    assert va.preferred_cmap == viridis_cmap

    # None is allowed - it is the default
    va = VisualAttributes(preferred_cmap=None)
    assert va.preferred_cmap is None



from echo import CallbackProperty, SelectionCallbackProperty
from glue.core.data_combo_helper import ComponentIDComboHelper

from glue.viewers.common.state import ViewerState


__all__ = ['TableViewerState']


class TableViewerState(ViewerState):
    """
    A state class that includes all the attributes for a table viewer.
    """

    filter_att = SelectionCallbackProperty(docstring='The component/column to filter/search on', default_index=0)
    filter = CallbackProperty(docstring='The text string to filter/search on')
    regex = CallbackProperty(docstring='Whether to apply regex to filter/search', default=False)

    def __init__(self, **kwargs):

        super(TableViewerState, self).__init__()
        self.filter_att_helper = ComponentIDComboHelper(self, 'filter_att', categorical=True, numeric=False)
        self.add_callback('layers', self._layers_changed)

        self.update_from_dict(kwargs)

    def _layers_changed(self, *args):

        layers_data = self.layers_data

        layers_data_cache = getattr(self, '_layers_data_cache', [])

        if layers_data == layers_data_cache:
            return

        self.filter_att_helper.set_multiple_data(self.layers_data)

        self._layers_data_cache = layers_data

    def _update_priority(self, name):
        if name == 'layers':
            return 2
        else:
            return 1



from textwrap import indent

from collections import defaultdict

import numpy as np

from glue.core import Subset
from echo import (delay_callback, CallbackProperty,
                                HasCallbackProperties, CallbackList)
from glue.core.state import saver, loader
from glue.core.component_id import PixelComponentID
from glue.core.exceptions import IncompatibleAttribute

__all__ = ['State', 'StateAttributeCacheHelper',
           'StateAttributeLimitsHelper', 'StateAttributeSingleValueHelper', 'StateAttributeHistogramHelper']


@saver(CallbackList)
def _save_callback_list(items, context):
    return {'values': [context.id(item) for item in items]}


@loader(CallbackList)
def _load_callback_list(rec, context):
    return [context.object(obj) for obj in rec['values']]


class State(HasCallbackProperties):
    """
    A class to represent the state of a UI element. Initially this doesn't add
    anything compared to HasCallbackProperties, but functionality will be added
    over time.
    """

    def __init__(self, **kwargs):
        super(State, self).__init__()
        self.update_from_dict(kwargs)

    def update_from_state(self, state):
        """
        Update this state using the values from another state.

        Parameters
        ----------
        state : `~glue.core.state_objects.State`
            The state to use the values from
        """
        self.update_from_dict(state.as_dict())

    def update_from_dict(self, properties):
        """
        Update this state using the values from a dictionary of attributes.

        Parameters
        ----------
        properties : dict
            The dictionary containing attribute/value pairs.
        """

        if len(properties) == 0:
            return

        # Group properties into priority so as to be able to update them in
        # chunks and not fire off callback events between every single one.
        groups = defaultdict(list)
        for name in properties:
            if self.is_callback_property(name):
                groups[self._update_priority(name)].append(name)

        for priority in sorted(groups, reverse=True):
            with delay_callback(self, *groups[priority]):
                for name in groups[priority]:
                    setattr(self, name, properties[name])

    def as_dict(self):
        """
        Return the current state as a dictionary of attribute/value pairs.
        """
        properties = {}
        for name in dir(self):
            if not name.startswith('_') and self.is_callback_property(name):
                properties[name] = getattr(self, name)
        return properties

    def __str__(self):
        s = ""
        state_dict = self.as_dict()
        for key in sorted(state_dict):
            value = state_dict[key]
            if np.isscalar(value):
                s += "{0}: {1}\n".format(key, value)
            else:
                s += "{0}: {1}\n".format(key, repr(value))
        return s.strip()

    def __repr__(self):
        return "<{0}\n{1}\n>".format(self.__class__.__name__, indent(str(self), '  '))

    def __gluestate__(self, context):
        return {'values': dict((key, context.id(value)) for key, value in self.as_dict().items())}

    def _update_priority(self, name):
        return 0

    @classmethod
    def __setgluestate__(cls, rec, context):
        properties = dict((key, context.object(value)) for key, value in rec['values'].items())
        result = cls(**properties)
        return result


class StateAttributeCacheHelper(object):
    """
    Generic class to help with caching values on a per-attribute basis

    Parameters
    ----------
    state : :class:`glue.core.state_objects.State`
        The state object with the callback properties to cache
    attribute : str
        The attribute name - this will be populated once a dataset is assigned
        to the helper
    cache : dict, optional
        A dictionary that can be used to hold the cache. This option can be used
        if a common cache should be shared between different helpers.
    kwargs
        Additional keyword arguments are taken to be values that should be
        used/cached. The key should be the name to be understood by sub-classes
        of this base class, and the value should be the name of the attribute
        in the state.
    """

    def __init__(self, state, attribute, cache=None, **kwargs):

        self._state = state
        self._attribute = attribute
        self._values = dict((key, kwargs[key]) for key in self.values_names if key in kwargs)
        self._modifiers = dict((key, kwargs[key]) for key in self.modifiers_names if key in kwargs)

        self._attribute_lookup = {'attribute': self._attribute}
        self._attribute_lookup.update(self._values)
        self._attribute_lookup.update(self._modifiers)
        self._attribute_lookup_inv = {v: k for k, v in self._attribute_lookup.items()}

        self._state.add_callback(self._attribute, self._update_attribute)

        self._state.add_global_callback(self._update_values)

        # NOTE: don't use self._cache = cache or {} here since if the initial
        #       cache is empty it will evaluate as False!
        if cache is None:
            self._cache = {}
        else:
            self._cache = cache

    @property
    def data_values(self):
        return self.data[self.component_id]

    def invalidate_cache(self):
        self._cache.clear()

    @property
    def data(self):
        if self.attribute is None:
            return None
        else:
            # For subsets in 'data' mode, we want to compute the limits based on
            # the full dataset, not just the subset.
            if isinstance(self.attribute.parent, Subset):
                return self.attribute.parent.data
            else:
                return self.attribute.parent

    @property
    def component_id(self):
        if self.attribute is None:
            return None
        else:
            return self.attribute

    def set_cache(self, cache):
        self._cache = cache
        self._update_attribute()

    def _update_attribute(self, *args):
        if self.component_id in self._cache:
            # The component ID already exists in the cache, so just revert to
            # that version of the values/settings.
            self.set(cache=False, **self._cache[self.component_id])
        else:
            # We need to compute the values for the first time
            self.update_values(attribute=self.component_id, use_default_modifiers=True)

    def _update_values(self, **properties):

        if hasattr(self, '_in_set'):
            if self._in_set:
                return
        if self.attribute is None:
            return
        properties = dict((self._attribute_lookup_inv[key], value)
                          for key, value in properties.items() if key in self._attribute_lookup_inv and self._attribute_lookup_inv[key] != 'attribute')
        if len(properties) > 0:
            self.update_values(**properties)

    def _modifiers_as_dict(self):
        return dict((prop, getattr(self, prop)) for prop in self.modifiers_names if prop in self._modifiers)

    def _values_as_dict(self):
        return dict((prop, getattr(self, prop)) for prop in self.values_names if prop in self._values)

    def _update_cache(self):
        if self.component_id is not None:
            self._cache[self.component_id] = {}
            self._cache[self.component_id].update(self._modifiers_as_dict())
            self._cache[self.component_id].update(self._values_as_dict())

    def __getattr__(self, attribute):
        if attribute in self._attribute_lookup:
            return getattr(self._state, self._attribute_lookup[attribute])
        else:
            raise AttributeError(attribute)

    def __setattr__(self, attribute, value):
        if attribute.startswith('_') or attribute not in self._attribute_lookup:
            return object.__setattr__(self, attribute, value)
        else:
            return setattr(self._state, self._attribute_lookup[attribute], value)

    def set(self, cache=True, **kwargs):

        self._in_set = True

        extra_kwargs = set(kwargs.keys()) - set(self.values_names) - set(self.modifiers_names)

        if len(extra_kwargs) > 0:
            raise ValueError("Invalid properties: {0}".format(extra_kwargs))

        with delay_callback(self._state, *self._attribute_lookup.values()):
            for prop, value in kwargs.items():
                setattr(self, prop, value)

        if cache:
            self._update_cache()

        self._in_set = False


class StateAttributeLimitsHelper(StateAttributeCacheHelper):
    """
    This class is a helper for attribute-dependent min/max level values. It
    is equivalent to AttributeLimitsHelper but operates on State objects and
    is GUI-independent.

    Parameters
    ----------
    attribute : str
        The attribute name - this will be populated once a dataset is assigned
        to the helper.
    random_subset : int
        How many points to use at most for the percentile calculation (using all
        values is highly inefficient and not needed)
    margin : float
        Whether to add a margin to the range of values determined. This should be
        given as a floating point value that will be multiplied by the range of
        values to give the margin to add to the limits.
    lower, upper : str
        The fields for the lower/upper levels
    percentile : ``QComboBox`` instance, optional
        The scale mode combo - this will be populated by presets such as
        Min/Max, various percentile levels, and Custom.
    log : bool
        Whether the limits are in log mode (in which case only positive values
        are used when finding the limits)

    Notes
    -----

    Once the helper is instantiated, the data associated with the helper can be
    set/changed with:

    >>> helper = AttributeLimitsHelper(...)
    >>> helper.data = data

    The data can also be passed to the initializer as described in the list of
    parameters above.
    """

    values_names = ('lower', 'upper')
    modifiers_names = ('log', 'percentile')

    def __init__(self, state, attribute, random_subset=10000, margin=0, **kwargs):

        super(StateAttributeLimitsHelper, self).__init__(state, attribute, **kwargs)

        self.margin = margin
        self.random_subset = random_subset
        self.subset_indices = None

        if self.attribute is not None:

            if (self.lower is not None and self.upper is not None and getattr(self, 'percentile', None) is None):
                # If the lower and upper limits are already set, we need to make
                # sure we don't override them, so we set the percentile mode to
                # custom if it isn't already set.
                self.set(percentile='Custom')
            else:
                # Otherwise, we force the recalculation or the fetching from
                # cache of the limits based on the current attribute
                self._update_attribute()

    def update_values(self, force=False, use_default_modifiers=False, **properties):

        if not force and not any(prop in properties for prop in ('attribute', 'percentile', 'log')):
            self.set(percentile='Custom')
            return

        if use_default_modifiers:
            percentile = 100
            log = False
        else:
            percentile = getattr(self, 'percentile', None) or 100
            log = getattr(self, 'log', None) or False

        if not force and (percentile == 'Custom' or not hasattr(self, 'data') or self.data is None):

            self.set(percentile=percentile, log=log)

        else:

            # Shortcut if the component ID is a pixel component ID
            if isinstance(self.component_id, PixelComponentID) and percentile == 100 and not log:
                lower = -0.5
                upper = self.data.shape[self.component_id.axis] - 0.5
                self.set(lower=lower, upper=upper, percentile=percentile, log=log)
                return

            exclude = (100 - percentile) / 2.

            if percentile == 100:
                lower = self.data.compute_statistic('minimum', cid=self.component_id,
                                                    finite=True, positive=log,
                                                    random_subset=self.random_subset)
                upper = self.data.compute_statistic('maximum', cid=self.component_id,
                                                    finite=True, positive=log,
                                                    random_subset=self.random_subset)
            else:
                lower = self.data.compute_statistic('percentile', cid=self.component_id,
                                                    percentile=exclude, positive=log,
                                                    random_subset=self.random_subset)
                upper = self.data.compute_statistic('percentile', cid=self.component_id,
                                                    percentile=100 - exclude, positive=log,
                                                    random_subset=self.random_subset)

            if not isinstance(lower, np.datetime64) and np.isnan(lower):
                lower, upper = 0, 1
            else:

                if self.data.get_kind(self.component_id) == 'categorical':
                    lower = np.floor(lower - 0.5) + 0.5
                    upper = np.ceil(upper + 0.5) - 0.5

                if log:
                    value_range = np.log10(upper / lower)
                    lower /= 10.**(value_range * self.margin)
                    upper *= 10.**(value_range * self.margin)
                else:
                    value_range = upper - lower
                    lower -= value_range * self.margin
                    upper += value_range * self.margin

            self.set(lower=lower, upper=upper, percentile=percentile, log=log)

    def flip_limits(self):
        self.set(lower=self.upper, upper=self.lower)


class StateAttributeSingleValueHelper(StateAttributeCacheHelper):

    values_names = ('value',)
    modifiers_names = ()

    def __init__(self, state, attribute, function, mode='values', **kwargs):
        self._function = function
        super(StateAttributeSingleValueHelper, self).__init__(state, attribute, **kwargs)
        if self.attribute is not None:
            self._update_attribute()
        if mode in ('values', 'component'):
            self.mode = mode
        else:
            raise ValueError('mode should be one of "values" or "component"')

    def update_values(self, use_default_modifiers=False, **properties):
        if not any(prop in properties for prop in ('attribute',)) or self.data is None:
            self.set()
        else:
            if self.mode == 'values':
                arg = self.data_values
            else:
                arg = self.data.get_component(self.component_id)
            self.set(value=self._function(arg))


class StateAttributeHistogramHelper(StateAttributeCacheHelper):

    values_names = ('lower', 'upper', 'n_bin')
    modifiers_names = ()

    def __init__(self, state, attribute, random_subset=10000, max_n_bin=30,
                 default_n_bin=15, common_n_bin=None, **kwargs):

        self._max_n_bin = max_n_bin
        self._default_n_bin = default_n_bin

        common_n_bin_att = common_n_bin

        super(StateAttributeHistogramHelper, self).__init__(state, attribute, **kwargs)

        self.random_subset = random_subset

        if common_n_bin_att is not None:
            if getattr(self._state, common_n_bin_att):
                self._common_n_bin = self._default_n_bin
            else:
                self._common_n_bin = None
            self._state.add_callback(common_n_bin_att, self._update_common_n_bin)
        else:
            self._common_n_bin = None

    def _apply_common_n_bin(self):
        for att in list(self._cache):
            try:
                if not self.data.get_kind(att) == 'categorical':
                    self._cache[att]['n_bin'] = self._common_n_bin
            except IncompatibleAttribute:
                # This can indicate that a dataset has been removed from the
                # data collection or that the attribute has changed to a
                # new dataset that is not compatible with the previous one.
                # In this case we should remove the entry
                self._cache.pop(att)

    def _update_common_n_bin(self, common_n_bin):
        if common_n_bin:
            if self.data.get_kind(self.component_id) == 'categorical':
                self._common_n_bin = self._default_n_bin
            else:
                self._common_n_bin = self.n_bin
            self._apply_common_n_bin()
        else:
            self._common_n_bin = None

    def update_values(self, force=False, use_default_modifiers=False, **properties):

        if not force and not any(prop in properties for prop in ('attribute', 'n_bin')) or self.data is None:
            self.set()
            return

        categorical = self.data.get_kind(self.component_id) == 'categorical'

        if 'n_bin' in properties:
            self.set()
            if self._common_n_bin is not None and not categorical:
                self._common_n_bin = properties['n_bin']
                self._apply_common_n_bin()

        if 'attribute' in properties or force:

            if categorical:

                n_categories = self.data_values.categories.size

                n_bin = max(1, min(n_categories, self._max_n_bin))
                lower = -0.5
                upper = lower + n_categories

            else:

                if self._common_n_bin is None:
                    n_bin = self._default_n_bin
                else:
                    n_bin = self._common_n_bin

                lower = self.data.compute_statistic('minimum', cid=self.component_id,
                                                    finite=True, random_subset=self.random_subset)
                upper = self.data.compute_statistic('maximum', cid=self.component_id,
                                                    finite=True, random_subset=self.random_subset)

                if not isinstance(lower, np.datetime64) and np.isnan(lower):
                    lower, upper = 0, 1

            self.set(lower=lower, upper=upper, n_bin=n_bin)


if __name__ == "__main__":

    from glue.core import Data

    class TestState(object):

        layer = CallbackProperty()
        comp = CallbackProperty(1)
        lower = CallbackProperty()
        higher = CallbackProperty()
        log = CallbackProperty()
        scale = CallbackProperty()

    state = TestState()
    state.layer = Data(x=np.arange(10),
                       y=np.arange(10) / 3.,
                       z=np.arange(10) - 5)

    helper = StateAttributeLimitsHelper(state, 'layer', 'comp',
                                        'lower', 'higher',
                                        percentile='scale', log='log')

    helper.component_id = state.layer.id['x']



"""
These classes define the behavior of how new subset states affect the
edit_subset of a Data object.
"""

import logging

from glue.core.contracts import contract
from glue.core.message import EditSubsetMessage
from glue.core.data_collection import DataCollection
from glue.core.data import Data
from glue.utils import as_list

__all__ = ['EditSubsetMode', 'NewMode', 'ReplaceMode', 'AndMode', 'OrMode',
           'XorMode', 'AndNotMode']


class EditSubsetMode(object):
    """
    Implements how new SubsetStates modify the edit_subset state
    """

    def __init__(self):
        self.data_collection = None
        self._mode = ReplaceMode
        self._edit_subset = []

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        if value is not self._mode:
            self._mode = value
            self._broadcast()

    @property
    def edit_subset(self):
        return self._edit_subset

    @edit_subset.setter
    def edit_subset(self, value):
        if value is self._edit_subset:
            return
        elif value is None:
            value = []
        self._edit_subset = value
        self._broadcast()

    def _broadcast(self):
        if self.data_collection is not None:
            self.data_collection.hub.broadcast(EditSubsetMessage(self, self.edit_subset, self.mode))

    def _combine_data(self, new_state, override_mode=None):
        """Dispatches to the combine method of mode attribute.

        The behavior is dependent on the mode it dispatches to.
        By default, the method uses ReplaceMode, which overwrites
        the edit_subsets' subset_state with new_state

        Parameters
        ----------
        new_state : :class:`~glue.core.subset.SubsetState`
            The new SubsetState.
        override_mode : callable, optional
            Mode to use instead of EditSubsetMode.mode.
        """
        mode = override_mode or self.mode
        if not self._edit_subset or mode is NewMode:
            if self.data_collection is None:
                raise RuntimeError("Must set data_collection before "
                                   "calling update")
            self.edit_subset = [self.data_collection.new_subset_group()]
        subs = self._edit_subset
        for s in as_list(subs):
            mode(s, new_state)

    @contract(d='inst($DataCollection, $Data)',
              new_state='isinstance(SubsetState)',
              focus_data='inst($Data)|None')
    def update(self, d, new_state, focus_data=None, override_mode=None):
        """ Apply a new subset state to editable subsets within a
        :class:`~glue.core.data.Data` or
        :class:`~glue.core.data_collection.DataCollection` instance

        Parameters
        ----------
        d : :class:`~glue.core.data.Data` or :class:`~glue.core.data_collection.DataCollection`
            The Data or Collection to act upon.
        new_state : :class:`~glue.core.subset.SubsetState`
            Subset state to combine with.
        focus_data : :class:`~glue.core.data.Data`, optional
            The main data set in focus by the client, if relevant.
            If a data set is in focus and has no subsets,
            a new one will be created using ``new_state``.
        override_mode : callable, optional
            Mode to use instead of ``EditSubsetMode.mode``.
        """
        logging.getLogger(__name__).debug("Update subset for %s", d)

        if isinstance(d, (Data, DataCollection)):
            self._combine_data(new_state, override_mode=override_mode)
        else:
            raise TypeError("input must be a Data or DataCollection: %s" %
                            type(d))


def NewMode(edit_subset, new_state):
    """ Replaces edit_subset.subset_state with new_state """
    logging.getLogger(__name__).debug("New %s", edit_subset)
    edit_subset.subset_state = new_state.copy()


def ReplaceMode(edit_subset, new_state):
    """ Replaces edit_subset.subset_state with new_state """
    logging.getLogger(__name__).debug("Replace %s", edit_subset)
    edit_subset.subset_state = new_state.copy()


def AndMode(edit_subset, new_state):
    """ Edit_subset.subset state is and-combined with new_state """
    new_state.parent = edit_subset
    state = new_state & edit_subset.subset_state
    edit_subset.subset_state = state


def OrMode(edit_subset, new_state):
    """ Edit_subset.subset state is or-combined with new_state """
    new_state.parent = edit_subset
    state = new_state | edit_subset.subset_state
    edit_subset.subset_state = state


def XorMode(edit_subset, new_state):
    """ Edit_subset.subset state is xor-combined with new_state """
    new_state.parent = edit_subset
    state = new_state ^ edit_subset.subset_state
    edit_subset.subset_state = state


def AndNotMode(edit_subset, new_state):
    """ Edit_subset.subset state is and-not-combined with new_state """
    new_state.parent = edit_subset
    state = edit_subset.subset_state & (~new_state)
    edit_subset.subset_state = state



"""
Classes to perform aggregations over cubes
"""

import numpy as np


def mom1(data, axis=0):
    """
    Intensity-weighted coordinate (function version). Pixel units.
    """

    shp = list(data.shape)
    n = shp.pop(axis)

    result = np.zeros(shp)
    w = np.zeros(shp)

    # build up slice-by-slice, to avoid big temporary cubes
    for loc in range(n):
        slc = tuple(loc if j == axis else slice(None) for j in range(data.ndim))
        val = data[slc]
        val = np.maximum(val, 0)
        result += val * loc
        w += val
    return result / w


def mom2(data, axis=0):
    """
    Intensity-weighted coordinate dispersion (function version). Pixel units.
    """

    shp = list(data.shape)
    n = shp.pop(axis)

    x = np.zeros(shp)
    x2 = np.zeros(shp)
    w = np.zeros(shp)

    # build up slice-by-slice, to avoid big temporary cubes
    for loc in range(n):
        slc = tuple(loc if j == axis else slice(None) for j in range(data.ndim))
        val = data[slc]
        val = np.maximum(val, 0)
        x += val * loc
        x2 += val * loc * loc
        w += val
    return np.sqrt(x2 / w - (x / w) ** 2)



import abc
import uuid
import logging

import numpy as np

from astropy.units import Quantity
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseLowLevelWCS

__all__ = ['Coordinates', 'IdentityCoordinates', 'AffineCoordinates',
           'coordinates_from_header', 'coordinates_from_wcs', 'header_from_string']


def default_world_coords(wcs):
    if isinstance(wcs, WCS):
        return wcs.wcs.crval
    else:
        return np.zeros(wcs.world_n_dim, dtype=float)


# The Coordinates class (and sub-classes) are used to map from pixel to world
# coordinates (and vice-versa). Rather than re-invent a new API, we adopt the
# one from the Astropy project which is generic and not specific to astronomy,
# which then allows us to make use of visualization tools (e.g. WCSAxes) built
# around this API. The API is defined/described in
# https://github.com/astropy/astropy-APEs/blob/master/APE14.rst


class Coordinates(BaseLowLevelWCS, metaclass=abc.ABCMeta):
    """
    Base class for coordinate transformation
    """

    def __init__(self, n_dim=None, pixel_n_dim=None, world_n_dim=None):
        self._pixel_n_dim = n_dim or pixel_n_dim
        self._world_n_dim = n_dim or world_n_dim
        self._world_uuids = [str(uuid.uuid4()) for i in range(self._world_n_dim)]

    @property
    def pixel_n_dim(self):
        return self._pixel_n_dim

    @property
    def world_n_dim(self):
        return self._world_n_dim

    @property
    def world_axis_physical_types(self):
        return [None] * self._world_n_dim

    @property
    def world_axis_units(self):
        return [''] * self._world_n_dim

    def array_index_to_world_values(self, *index_arrays):
        return self.pixel_to_world_values(*index_arrays[::-1])

    def world_to_array_index_values(self, *world_arrays):
        # Compatibility code, remove once we require Astropy 4.1 or later
        pixel_arrays = self.world_to_pixel_values(*world_arrays)
        if self.pixel_n_dim == 1:
            pixel_arrays = (pixel_arrays,)
        else:
            pixel_arrays = pixel_arrays[::-1]
        array_indices = tuple(np.asarray(np.floor(pixel + 0.5), dtype=np.int_) for pixel in pixel_arrays)
        return array_indices[0] if self.pixel_n_dim == 1 else array_indices

    @property
    def world_axis_object_components(self):
        return [(uid, 0, 'value') for uid in self._world_uuids]

    @property
    def world_axis_object_classes(self):
        return {self._world_uuids[idx]: (Quantity, (),
                                         {'unit': self.world_axis_units[idx]})
                for idx in range(self._world_n_dim)}

    def __gluestate__(self, context):
        return {}  # no state

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls()


class LegacyCoordinates(Coordinates):

    def __init__(self):
        super().__init__(pixel_n_dim=10, world_n_dim=10)

    def pixel_to_world_values(self, *pixel):
        if len(pixel) == 1:
            return pixel[0]
        else:
            return pixel

    def world_to_pixel_values(self, *world):
        if len(world) == 1:
            return world[0]
        else:
            return world


class IdentityCoordinates(Coordinates):

    def __init__(self, n_dim=None):
        super().__init__(pixel_n_dim=n_dim, world_n_dim=n_dim)

    def pixel_to_world_values(self, *pixel):
        if self.pixel_n_dim == 1:
            return pixel[0]
        else:
            return pixel

    def world_to_pixel_values(self, *world):
        if self.world_n_dim == 1:
            return world[0]
        else:
            return world

    @property
    def axis_correlation_matrix(self):
        return np.identity(self.pixel_n_dim).astype(bool)

    def __gluestate__(self, context):
        return {'ndim': self.pixel_n_dim}

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(n_dim=rec['ndim'])


class AffineCoordinates(Coordinates):
    """
    Coordinates determined via an affine transformation represented by an
    augmented matrix of shape N+1 x N+1 matrix, where N is the number of pixel
    and world coordinates. The last column of the matrix should be used for
    the translation term, and the last row should be set to 0 except for the
    last column which should be 1.

    Note that the order of the dimensions in the matrix (x, y) should be the
    opposite of the order of the order of dimensions of Numpy arrays (y, x).
    """

    def __init__(self, matrix, units=None, labels=None):

        if matrix.ndim != 2:
            raise ValueError("Affine matrix should be two-dimensional")

        if matrix.shape[0] != matrix.shape[1]:
            raise ValueError("Affine matrix should be square")

        if np.any(matrix[-1, :-1] != 0) or matrix[-1, -1] != 1:
            raise ValueError("Last row of matrix should be zeros and a one")

        if units is not None and len(units) != matrix.shape[0] - 1:
            raise ValueError("Expected {0} units, got {1}".format(matrix.shape[0] - 1, len(units)))

        if labels is not None and len(labels) != matrix.shape[0] - 1:
            raise ValueError("Expected {0} labels, got {1}".format(matrix.shape[0] - 1, len(labels)))

        self._matrix = matrix
        self._matrix_inv = np.linalg.inv(matrix)
        self._units = units
        self._labels = labels

        super().__init__(pixel_n_dim=self._matrix.shape[0] - 1,
                         world_n_dim=self._matrix.shape[0] - 1)

    def pixel_to_world_values(self, *pixel):
        scalar = np.all([np.isscalar(p) for p in pixel])
        pixel = np.array(np.broadcast_arrays(*(list(pixel) + [np.ones(np.shape(pixel[0]))])))
        pixel = np.moveaxis(pixel, 0, -1)
        world = np.matmul(pixel, self._matrix.T)
        world = tuple(np.moveaxis(world, -1, 0))[:-1]
        if self._matrix.shape[0] == 2:  # 1D
            return world[0]
        else:
            return world

    def world_to_pixel_values(self, *world):
        scalar = np.all([np.isscalar(w) for w in world])
        world = np.array(np.broadcast_arrays(*(list(world) + [np.ones(np.shape(world[0]))])))
        world = np.moveaxis(world, 0, -1)
        pixel = np.matmul(world, self._matrix_inv.T)
        pixel = tuple(np.moveaxis(pixel, -1, 0))[:-1]
        if self._matrix.shape[0] == 2:  # 1D
            return pixel[0]
        else:
            return pixel

    @property
    def world_axis_names(self):
        return self._labels or [''] * self.world_n_dim

    @property
    def world_axis_units(self):
        return self._units or [''] * self.world_n_dim

    def __gluestate__(self, context):
        return dict(matrix=context.do(self._matrix),
                    labels=self._labels,
                    units=self._units)

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(context.object(rec['matrix']),
                   units=rec['units'],
                   labels=rec['labels'])


# Kept for backward-compatibility
WCSCoordinates = WCS


def coordinates_from_header(header, hdulist=None):
    """
    Convert a FITS header into a glue Coordinates object.

    Parameters
    ----------
    header : :class:`astropy.io.fits.Header`
        Header to convert

    hdulist : :class:`astropy.io.fits.HDUList`, optional
        The full HDUList is required when header keywords point to a
        'Distortion Paper' lookup table stored in a different extension.

    Returns
    -------
    coordinates : :class:`~glue.core.coordinates.Coordinates`
    """

    # We check whether the header contains at least CRVAL1 - if not, we would
    # end up with a default WCS that isn't quite 1 to 1 (because of a 1-pixel
    # offset) so better use Coordinates in that case.

    from astropy.io.fits import Header

    if isinstance(header, Header) and 'CRVAL1' in header:
        try:
            return WCSCoordinates(header, hdulist)
        except Exception as e:
            logging.getLogger(__name__).warn(
                "\n\n*******************************\n"
                "Encounted an error during WCS parsing. "
                "Discarding world coordinates! "
                "\n{}\n"
                "*******************************\n\n".format(str(e)))
    try:
        return IdentityCoordinates(header.get('NAXIS', 2))
    except Exception:
        return IdentityCoordinates(2)


def coordinates_from_wcs(wcs):
    """
    Convert an Astropy WCS object into a glue Coordinates object.

    Parameters
    ----------
    wcs : :class:`astropy.wcs.WCS`
        The WCS object to use

    Returns
    -------
    coordinates : :class:`~glue.core.coordinates.Coordinates`
    """
    return WCSCoordinates(wcs.to_header())


def header_from_string(string):
    """
    Convert a string to a FITS header.
    """
    from astropy.io import fits
    return fits.Header.fromstring(string, sep='\n')



from glue.core.data import Data, Component
from glue.config import data_factory
from glue.core.data_factories.helpers import has_extension

__all__ = ['is_npy_npz', 'npy_npz_reader']


def is_npy_npz(filename):
    """
    The first bytes are x93NUMPY (for npy) or PKx03x04 (for npz)
    """
    # See: https://github.com/numpy/numpy/blob/master/doc/neps/npy-format.rst
    from numpy.lib.format import MAGIC_PREFIX
    MAGIC_PREFIX_NPZ = b'PK\x03\x04'  # first 4 bytes for a zipfile
    tester = has_extension('npz .npz')
    with open(filename, 'rb') as infile:
        prefix = infile.read(6)
    return prefix == MAGIC_PREFIX or (tester(filename) and prefix[:4] == MAGIC_PREFIX_NPZ)


@data_factory(label="Numpy save file", identifier=is_npy_npz, priority=100)
def npy_npz_reader(filename, format='auto', auto_merge=False, **kwargs):
    """
    Read in a Numpy structured array saved to a .npy or .npz file.

    Parameters
    ----------
    source: str
        The pathname to the Numpy save file.
    """

    import numpy as np
    data = np.load(filename)

    if isinstance(data, np.ndarray):
        data = {None: data}

    groups = []
    for groupname in sorted(data):

        d = Data(label=groupname)
        arr = data[groupname]

        if arr.dtype.names is None:
            comp = Component.autotyped(arr)
            d.add_component(comp, label='array')
        else:
            for name in arr.dtype.names:
                comp = Component.autotyped(arr[name])
                d.add_component(comp, label=name)

        groups.append(d)

    return groups



import os
import warnings

from IPython import get_ipython

from glue.core.hub import HubListener
from glue.core import BaseData, Subset
from glue.core import command
from glue.core.command import ApplySubsetState

from glue.core.state import save
from glue.core import message as msg
from glue.core.exceptions import IncompatibleDataException
from glue.core.state import lookup_class_with_patches
from echo import delay_callback, ignore_callback
from glue.core.layer_artist import LayerArtistContainer

from glue.viewers.common.state import ViewerState
from glue.viewers.common.layer_artist import LayerArtist

from glue.config import layer_artist_maker

__all__ = ['BaseViewer', 'Viewer']


def get_layer_artist_from_registry(data, viewer):
    """
    Check whether any plugins define an appropriate custom layer artist for
    the specified data and viewer.
    """
    for maker in layer_artist_maker.members:
        layer_artist = maker.function(viewer, data)
        if layer_artist is not None:
            return layer_artist


class BaseViewer(HubListener):
    """
    The base class for all viewers.
    """

    LABEL = None

    def __init__(self, session):
        self._session = session
        self._data = session.data_collection
        self._hub = None
        if self.LABEL is None:
            self.LABEL = str(self.__class__)

    def register_to_hub(self, hub):
        self._hub = hub

    @property
    def session(self):
        return self._session

    def request_add_layer(self, layer):
        """ Issue a command to add a layer """
        cmd = command.AddLayer(layer=layer, viewer=self)
        self._session.command_stack.do(cmd)

    def add_layer(self, layer):
        if isinstance(layer, BaseData):
            self.add_data(layer)
        elif isinstance(layer, Subset):
            self.add_subset(layer)

    def remove_layer(self, layer):
        pass

    def add_data(self, data):
        raise NotImplementedError()

    def add_subset(self, subset):
        raise NotImplementedError()

    def __str__(self):
        return self.LABEL

    def __gluestate__(self, context):
        return dict(session=context.id(self._session))

    @classmethod
    def __setgluestate__(cls, rec, context):
        session = context.object(rec['session'])
        return cls(session)

    def apply_subset_state(self, subset_state, override_mode=None):
        cmd = ApplySubsetState(data_collection=self._data,
                               subset_state=subset_state,
                               override_mode=override_mode)
        self._session.command_stack.do(cmd)


TEMPLATE_SCRIPT = """
# This script was produced by glue and can be used to further customize a
# particular plot.

### Package imports

{imports}

### Set up data

data_collection = load('{data}')

### Set up viewer

{header}

### Set up layers

{layers}

### Legend

{legend}

### Finalize viewer

{footer}
""".strip()


class Viewer(BaseViewer):
    """
    A viewer class that uses a state class to represent the overall viewer
    state, and uses layer artists and state classes to handle each dataset
    and subset in the data viewer.
    """

    # The LayerArtistContainer class/subclass to use
    _layer_artist_container_cls = LayerArtistContainer

    # The state class/subclass to use
    _state_cls = ViewerState

    _data_artist_cls = LayerArtist
    _subset_artist_cls = LayerArtist

    allow_duplicate_data = False
    allow_duplicate_subset = False

    large_data_size = None

    def __init__(self, session, state=None):

        super(Viewer, self).__init__(session)

        # Set up the state which will contain everything needed to represent
        # the current state of the viewer
        self.state = state or self._state_cls()
        self.state.data_collection = session.data_collection

        # Create the layer artist container, which is the object in which
        # we will add LayerArtist objects
        self._layer_artist_container = self._layer_artist_container_cls()

        # When layer artists are removed from the layer artist container, we
        # need to make sure we remove matching layer states in the viewer state
        # layers attribute.
        self._layer_artist_container.on_changed(self._sync_state_layers)

        # And vice-versa when layer states are removed from the viewer state, we
        # need to keep the layer_artist_container in sync
        self.state.add_callback('layers', self._sync_layer_artist_container, priority=10000)

        self.state.add_callback('layers', self.draw_legend)

    def draw_legend(self, *args):
        pass

    def _sync_state_layers(self, *args):
        # Remove layer state objects that no longer have a matching layer
        for layer_state in self.state.layers:
            if layer_state.layer not in self._layer_artist_container:
                self.state.layers.remove(layer_state)

    def _sync_layer_artist_container(self, *args):
        # Remove layer artists that no longer have a matching layer state
        layer_states = set(layer_state.layer for layer_state in self.state.layers)
        for layer_artist in self._layer_artist_container:
            if layer_artist.layer not in layer_states:
                self._layer_artist_container.remove(layer_artist)

    def warn(self, message, *args, **kwargs):
        warnings.warn(message)
        return True

    def add_data(self, data):

        # Check if data already exists in viewer
        if not self.allow_duplicate_data and data in self._layer_artist_container:
            return True

        if self.large_data_size is not None and data.size >= self.large_data_size:
            proceed = self.warn('Add large data set?', 'Data set {0:s} has {1:d} points, and '
                                'may render slowly.'.format(data.label, data.size),
                                default='Cancel', setting='show_large_data_warning')
            if not proceed:
                return False

        if data not in self.session.data_collection:
            raise IncompatibleDataException("Data not in DataCollection")

        # Create layer artist and add to container. First check whether any
        # plugins want to make a custom layer artist.
        layer = get_layer_artist_from_registry(data, self) or self.get_data_layer_artist(data)

        if layer is None:
            return False

        # When adding a layer artist to the layer artist container, zorder
        # gets set automatically - however since we call a forced update of the
        # layer after adding it to the container we can ignore any callbacks
        # related to zorder. We also then need to set layer.state.zorder manually.
        with ignore_callback(layer.state, 'zorder'):
            self._layer_artist_container.append(layer)
        layer.update()
        self.draw_legend()  # need to be called here because callbacks are ignored in previous step

        # Add existing subsets to viewer
        for subset in data.subsets:
            self.add_subset(subset)

        return True

    def remove_data(self, data):
        with delay_callback(self.state, 'layers'):
            for layer_state in self.state.layers[::-1]:
                if isinstance(layer_state.layer, BaseData):
                    if layer_state.layer is data:
                        self.state.layers.remove(layer_state)
                else:
                    if layer_state.layer.data is data:
                        self.state.layers.remove(layer_state)

    def get_data_layer_artist(self, layer=None, layer_state=None):
        return self.get_layer_artist(self._data_artist_cls, layer=layer, layer_state=layer_state)

    def get_subset_layer_artist(self, layer=None, layer_state=None):
        return self.get_layer_artist(self._subset_artist_cls, layer=layer, layer_state=layer_state)

    def get_layer_artist(self, cls, layer=None, layer_state=None):
        return cls(self.state, layer=layer, layer_state=layer_state)

    def add_subset(self, subset):

        # Check if subset already exists in viewer
        if not self.allow_duplicate_subset and subset in self._layer_artist_container:
            return True

        # Create layer artist and add to container. First check whether any
        # plugins want to make a custom layer artist.
        layer = get_layer_artist_from_registry(subset, self) or self.get_subset_layer_artist(subset)

        if layer is None:
            return False

        # When adding a layer artist to the layer artist container, zorder
        # gets set automatically - however since we call a forced update of the
        # layer after adding it to the container we can ignore any callbacks
        # related to zorder.
        with ignore_callback(layer.state, 'zorder'):
            self._layer_artist_container.append(layer)
        layer.update()
        self.draw_legend()  # need to be called here because callbacks are ignored in previous step

        return True

    def remove_subset(self, subset):
        if subset in self._layer_artist_container:
            self._layer_artist_container.pop(subset)

    def _add_subset(self, message):
        self.add_subset(message.subset)

    def _update_data_numerical(self, message):
        # For some viewers, we might want to do additional things when the
        # actual numerical values or shape of a dataset change, but by default
        # we just pass this on to _update_data
        self._update_data(message)

    def _update_data(self, message):
        if message.data in self._layer_artist_container:
            for layer_artist in self._layer_artist_container:
                if isinstance(layer_artist.layer, Subset):
                    if layer_artist.layer.data is message.data:
                        layer_artist.update()
                        try:
                            components_changed = message.components_changed
                            layer_artist._on_components_changed(components_changed)
                        except AttributeError:
                            pass

                else:
                    if layer_artist.layer is message.data:
                        layer_artist.update()
                        try:
                            components_changed = message.components_changed
                            layer_artist._on_components_changed(components_changed)
                        except AttributeError:
                            pass

    def _update_subset(self, message):
        if message.attribute == 'style':
            return
        if message.subset in self._layer_artist_container:
            for layer_artist in self._layer_artist_container[message.subset]:
                layer_artist.update()

    def _remove_subset(self, message):
        self.remove_subset(message.subset)

    def options_widget(self):
        return self.options

    def _subset_has_data(self, x):
        return x.sender.data in self._layer_artist_container.layers

    def _has_data_or_subset(self, x):
        return x.sender in self._layer_artist_container.layers

    def _remove_data(self, message):
        self.remove_data(message.data)

    def _is_appearance_settings(self, msg):
        return ('BACKGROUND_COLOR' in msg.settings or
                'FOREGROUND_COLOR' in msg.settings)

    def register_to_hub(self, hub):

        super(Viewer, self).register_to_hub(hub)

        hub.subscribe(self, msg.SubsetCreateMessage,
                      handler=self._add_subset,
                      filter=self._subset_has_data)

        hub.subscribe(self, msg.SubsetUpdateMessage,
                      handler=self._update_subset,
                      filter=self._has_data_or_subset)

        hub.subscribe(self, msg.SubsetDeleteMessage,
                      handler=self._remove_subset,
                      filter=self._has_data_or_subset)

        hub.subscribe(self, msg.NumericalDataChangedMessage,
                      handler=self._update_data_numerical,
                      filter=self._has_data_or_subset)

        hub.subscribe(self, msg.DataCollectionDeleteMessage,
                      handler=self._remove_data)

        hub.subscribe(self, msg.ComponentsChangedMessage,
                      handler=self._update_data,
                      filter=self._has_data_or_subset)

        hub.subscribe(self, msg.ExternallyDerivableComponentsChangedMessage,
                      handler=self._update_data,
                      filter=self._has_data_or_subset)

        hub.subscribe(self, msg.SettingsChangeMessage,
                      self._update_appearance_from_settings,
                      filter=self._is_appearance_settings)

        hub.subscribe(self, msg.ComputationMessage,
                      self._update_computation,
                      filter=self._has_layer_artist)

        hub.subscribe(self, msg.LayerArtistDisabledMessage,
                      self.draw_legend,
                      filter=self._has_layer_artist)

    def _has_layer_artist(self, message):
        return message.layer_artist in self.layers

    def _update_computation(self, message=None):
        pass

    def _update_appearance_from_settings(self, message=None):
        pass

    def __str__(self):
        return self.state.title or self.LABEL

    def __gluestate__(self, context):
        return dict(state=self.state.__gluestate__(context),
                    session=context.id(self._session),
                    layers=list(map(context.do, self.layers)),
                    _protocol=1)

    @classmethod
    def __setgluestate__(cls, rec, context):

        session = context.object(rec['session'])
        viewer_state = cls._state_cls.__setgluestate__(rec['state'], context)

        viewer = cls(session, state=viewer_state)
        viewer.register_to_hub(session.hub)

        # Restore layer artists. Ideally we would delay instead of ignoring the
        # callback here.
        with viewer._layer_artist_container.ignore_callbacks():
            for l in rec['layers']:
                cls = lookup_class_with_patches(l.pop('_type'))
                layer_state = context.object(l['state'])
                layer_artist = viewer.get_layer_artist(cls, layer_state=layer_state)
                layer_state.viewer_state = viewer.state
                viewer._layer_artist_container.append(layer_artist)

        viewer.draw_legend()  # need to be called here because callbacks are ignored in previous step
        return viewer

    def cleanup(self):

        if self._hub is not None:
            self.unregister(self._hub)

        self._layer_artist_container.clear_callbacks()
        self._layer_artist_container.clear()

        # Remove any references to the viewer in the IPython namespace. We use
        # list() here to force an explicit copy since we are modifying the
        # dictionary in-place
        shell = get_ipython()
        if shell is not None:
            for key in list(shell.user_ns):
                if shell.user_ns[key] is self:
                    shell.user_ns.pop(key)

    def remove_layer(self, layer):
        self._layer_artist_container.pop(layer)

    @property
    def layers(self):
        """Return a tuple of layers in this viewer.

        A layer is a visual representation of a dataset or subset within
        the viewer"""
        return tuple(self._layer_artist_container)

    def _script_header(self):
        raise NotImplementedError()

    def _script_legend(self):
        return [], ""

    def _script_footer(self):
        raise NotImplementedError()

    def export_as_script(self, filename):

        data_filename = os.path.relpath(filename) + '.data'

        save(data_filename, self.session.data_collection)

        imports = ['from glue.core.state import load']

        imports_header, header = self._script_header()
        imports.extend(imports_header)

        layers = ""
        for ilayer, layer in enumerate(self.layers):
            if layer.layer.label:
                layers += '## Layer {0}: {1}\n\n'.format(ilayer + 1, layer.layer.label)
            else:
                layers += '## Layer {0}\n\n'.format(ilayer + 1)
            if layer.visible and layer.enabled:
                if isinstance(layer.layer, BaseData):
                    index = self.session.data_collection.index(layer.layer)
                    layers += "layer_data = data_collection[{0}]\n\n".format(index)
                else:
                    dindex = self.session.data_collection.index(layer.layer.data)
                    sindex = layer.layer.data.subsets.index(layer.layer)
                    layers += ("layer_data = data_collection[{0}].subsets[{1}]\n\n"
                               .format(dindex, sindex))
            imports_layer, layer_script = layer._python_exporter(layer)
            if layer_script is None:
                continue
            imports.extend(imports_layer)
            layers += layer_script.strip() + "\n"

        imports_legend, legend = self._script_legend()
        imports.extend(imports_legend)
        imports_footer, footer = self._script_footer()
        imports.extend(imports_footer)

        imports = os.linesep.join(sorted(set(imports),
                                         key=lambda s: s.strip('# ')))
        # The sorting key is added keep together similar but commented imports
        # Typical ex:
        #    matplotlib.use('Agg')
        #    # matplotlib.use('qt5Agg')

        script = TEMPLATE_SCRIPT.format(data=os.path.basename(data_filename),
                                        imports=imports.strip(),
                                        header=header.strip(),
                                        layers=layers.strip(),
                                        legend=legend.strip(),
                                        footer=footer.strip())

        with open(filename, 'w') as f:
            f.write(script)



import numpy as np
from echo import delay_callback
from glue.tests.visual.helpers import visual_test
from glue.viewers.image.viewer import SimpleImageViewer
from glue.core.application_base import Application
from glue.core.data import Data
from glue.core.link_helpers import LinkSame
from glue.core.data_region import RegionData

from shapely.geometry import Polygon, MultiPolygon


@visual_test
def test_simple_viewer():

    # Make sure the simple viewer can be instantiated

    data1 = Data(x=np.arange(6).reshape((2, 3)), label='data1')
    data2 = Data(y=2 * np.arange(6).reshape((2, 3)), label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.pixel_component_ids[1] > 1.2)

    return viewer.figure


@visual_test
def test_region_layer():
    poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
    poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
    poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
    poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

    polygons = MultiPolygon([poly_3, poly_4])

    geoms = np.array([poly_1, poly_2, polygons])
    values = np.array([1, 2, 3])
    region_data = RegionData(regions=geoms, values=values)

    image_data = Data(x=np.arange(10000).reshape((100, 100)), label='data1')
    app = Application()
    app.data_collection.append(image_data)
    app.data_collection.append(region_data)

    link1 = LinkSame(region_data.center_x_id, image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.center_y_id, image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(image_data)
    viewer.add_data(region_data)

    return viewer.figure


@visual_test
def test_region_layer_flip():
    poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
    poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
    poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
    poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

    polygons = MultiPolygon([poly_3, poly_4])

    geoms = np.array([poly_1, poly_2, polygons])
    values = np.array([1, 2, 3])
    region_data = RegionData(regions=geoms, values=values)

    image_data = Data(x=np.arange(10000).reshape((100, 100)), label='data1')
    app = Application()
    app.data_collection.append(image_data)
    app.data_collection.append(region_data)

    link1 = LinkSame(region_data.center_x_id, image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.center_y_id, image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(image_data)
    viewer.add_data(region_data)

    # We need this delay callback here because, while this works in the QT GUI,
    # we need to make sure not to try and redraw the regions while we are flipping
    # the coordinates displayed.

    with delay_callback(viewer.state, 'x_att', 'y_att'):
        viewer.state.x_att = image_data.pixel_component_ids[0]
        viewer.state.y_att = image_data.pixel_component_ids[1]

    return viewer.figure



from .command import Command, CommandStack  # noqa
from .component import Component  # noqa
from .component_id import ComponentID  # noqa
from .component_link import ComponentLink  # noqa
from .coordinates import Coordinates  # noqa  # noqa
from .data import BaseData, BaseCartesianData, Data  # noqa
from .data_collection import DataCollection  # noqa
from .hub import Hub, HubListener  # noqa
from .link_manager import LinkManager  # noqa
from .session import Session  # noqa
from .subset import Subset  # noqa
from .subset_group import SubsetGroup  # noqa
from .visual import VisualAttributes  # noqa

# We import this last to avoid circular imports
from . import parsers  # noqa
from .application_base import Application  # noqa



import sys

if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources


__all__ = ['icon_path']


def icon_path(icon_name, icon_format='png'):
    """
    Return the absolute path to an icon

    Parameters
    ----------
    icon_name : str
       Name of icon, without extension or directory prefix
    icon_format : str, optional
        Can be either 'png' or 'svg'

    Returns
    -------
    path : str
      Full path to icon
    """

    icon_name += '.{0}'.format(icon_format)

    icon_file = importlib_resources.files("glue") / "icons" / icon_name
    if icon_file.is_file():
        return str(icon_file)
    else:
        raise RuntimeError("Icon does not exist: %s" % icon_name)



import numpy as np
from numpy.testing import assert_allclose

from glue.core.component_id import ComponentID

from ..data_collection import DataCollection
from ..link_helpers import ComponentLink
from ..data import BaseCartesianData
from ..coordinates import IdentityCoordinates


class ExampleBaseData(BaseCartesianData):

    def __init__(self, cid_label='x', coords=None):
        super().__init__(coords=coords)
        self.data_cid = ComponentID(label=cid_label, parent=self)
        self._array = np.arange(12).reshape((1, 4, 3))

    @property
    def label(self):
        return "Example Data"

    @property
    def shape(self):
        return (1, 4, 3)

    @property
    def main_components(self):
        return [self.data_cid]

    def get_kind(self, cid):
        return 'numerical'

    def get_data(self, cid, view=None):
        if cid is self.data_cid:
            if view is None:
                return self._array
            else:
                return self._array[view]
        else:
            return super(ExampleBaseData, self).get_data(cid, view=view)

    def get_mask(self, subset_state, view=None):
        return subset_state.to_mask(self, view=view)

    def compute_statistic(self, statistic, cid,
                          axis=None, finite=True,
                          positive=False, subset_state=None,
                          percentile=None, random_subset=None):
        raise NotImplementedError()

    def compute_histogram(self, cid,
                          range=None, bins=None, log=False,
                          subset_state=None, subset_group=None):
        raise NotImplementedError()


def test_data_coords():

    # Make sure that world_component_ids works in both the case where
    # coords is not defined and when it is defined.

    data1 = ExampleBaseData()
    assert len(data1.pixel_component_ids) == 3
    assert len(data1.world_component_ids) == 0

    data2 = ExampleBaseData(coords=IdentityCoordinates(n_dim=3))
    assert len(data2.pixel_component_ids) == 3
    assert len(data2.world_component_ids) == 3

    for idx in range(3):

        assert_allclose(data2[data2.world_component_ids[idx]],
                        data2[data2.pixel_component_ids[idx]])

        assert_allclose(data2[data2.world_component_ids[idx], (0, slice(2), 2)],
                        data2[data2.pixel_component_ids[idx], (0, slice(2), 2)])


def test_linking():

    data1 = ExampleBaseData(cid_label='x')
    data2 = ExampleBaseData(cid_label='y')
    data3 = ExampleBaseData(cid_label='z')

    dc = DataCollection([data1, data2, data3])

    cid1 = data1.main_components[0]
    cid2 = data2.main_components[0]
    cid3 = data3.main_components[0]

    def double(x):
        return 2 * x

    def halve(x):
        return x / 2.

    link1 = ComponentLink([cid1], cid2, using=double, inverse=halve)
    link2 = ComponentLink([cid2], cid3, using=double, inverse=halve)

    dc.add_link(link1)
    dc.add_link(link2)

    assert_allclose(data1[cid2], data1[cid1] * 2)
    assert_allclose(data1[cid3], data1[cid1] * 4)

    assert_allclose(data2[cid1], data1[cid1] / 2)
    assert_allclose(data3[cid1], data1[cid1] / 4)

    assert_allclose(data2[cid2], data1[cid2] / 2)
    assert_allclose(data3[cid2], data1[cid2] / 4)

    assert_allclose(data2[cid3], data1[cid3] / 2)
    assert_allclose(data3[cid3], data1[cid3] / 4)


def test_pixel_aligned():

    data1 = ExampleBaseData(cid_label='x')
    data2 = ExampleBaseData(cid_label='y')
    data3 = ExampleBaseData(cid_label='z')

    dc = DataCollection([data1, data2, data3])

    def double(x):
        return 2 * x

    def halve(x):
        return x / 2.

    for dim in range(3):

        cid1 = data1.pixel_component_ids[dim]
        cid2 = data2.pixel_component_ids[dim]
        cid3 = data3.pixel_component_ids[dim]

        link1 = ComponentLink([cid1], cid2)

        if dim == 1:
            link2 = ComponentLink([cid2], cid3, using=double, inverse=halve)
        else:
            link2 = ComponentLink([cid2], cid3)

        dc.add_link(link1)
        dc.add_link(link2)

    assert len(data1.pixel_aligned_data) == 1
    assert data1.pixel_aligned_data[data2] == [0, 1, 2]
    assert len(data2.pixel_aligned_data) == 1
    assert data2.pixel_aligned_data[data1] == [0, 1, 2]
    assert len(data3.pixel_aligned_data) == 0



# Define decorators that can be used for pytest tests

import os
import zlib
import tempfile
from contextlib import contextmanager
from packaging.version import Version

import pytest


def make_marker(mark_creator, module, label=None, version=None, mark_if='lt'):
    label = label or module
    try:
        if label == 'PyQt5':  # PyQt5 does not use __version__
            from PyQt5 import QtCore
            version_installed = QtCore.PYQT_VERSION_STR
        elif label == 'PyQt6':  # PyQt6 does not use __version__
            from PyQt6 import QtCore
            version_installed = QtCore.PYQT_VERSION_STR
        else:
            mod = __import__(module)
            version_installed = mod.__version__
        if version:
            options = ['lt', 'le', 'gt', 'ge']
            if mark_if not in options:
                raise ValueError(f"mark_if must be one of {options}")
            assert not getattr(Version(version_installed), f'__{mark_if}__')(Version(version))
        installed = True
    except (ImportError, AssertionError, AttributeError):
        installed = False
    return installed, mark_creator(installed, label, version)


def make_skipper(module, label=None, version=None, skip_if='lt'):
    def mark_creator(installed, lbl, vrs):
        return pytest.mark.skipif(str(not installed), reason=f'Requires {lbl} not {skip_if} {vrs}')
    return make_marker(mark_creator, module, label=label, version=version, mark_if=skip_if)


def make_xfailer(module, label=None, version=None, xfail_if='lt'):
    def mark_creator(installed, lbl, vrs):
        return pytest.mark.xfail(condition=not installed, reason=f'Fails if {lbl} {xfail_if} {vrs}')
    return make_marker(mark_creator, module, label=label, version=version, mark_if=xfail_if)


ASTROPY_INSTALLED, requires_astropy = make_skipper('astropy', label='Astropy')

MATPLOTLIB_GE_22, requires_matplotlib_ge_22 = make_skipper('matplotlib', version='2.2')

MATPLOTLIB_GE_37, xfail_matplotlib_lt_37 = make_xfailer('matplotlib', version='3.7')

ASTRODENDRO_INSTALLED, requires_astrodendro = make_skipper('astrodendro')

SCIPY_INSTALLED, requires_scipy = make_skipper('scipy', label='SciPy')

PIL_INSTALLED, requires_pil = make_skipper('PIL', label='PIL')

SKIMAGE_INSTALLED, requires_skimage = make_skipper('skimage', label='scikit-image')

XLRD_INSTALLED, requires_xlrd = make_skipper('xlrd')

OPENPYXL_INSTALLED, requires_openpyxl = make_skipper('openpyxl')

PYAVM_INSTALLED, requires_pyavm = make_skipper('pyavm')

PLOTLY_INSTALLED, requires_plotly = make_skipper('plotly')

IPYTHON_INSTALLED, requires_ipython = make_skipper('IPython')

requires_pil_or_skimage = pytest.mark.skipif(str(not SKIMAGE_INSTALLED and not PIL_INSTALLED),
                                             reason='Requires PIL or scikit-image')

PLOTLY_INSTALLED, requires_plotly = make_skipper('plotly')

H5PY_INSTALLED, requires_h5py = make_skipper('h5py')

PYQT5_INSTALLED, requires_pyqt5 = make_skipper('PyQt5')
PYQT6_INSTALLED, requires_pyqt6 = make_skipper('PyQt6')
PYSIDE2_INSTALLED, requires_pyside2 = make_skipper('PySide2')
PYSIDE6_INSTALLED, requires_pyside6 = make_skipper('PySide6')

HYPOTHESIS_INSTALLED, requires_hypothesis = make_skipper('hypothesis')

QT_INSTALLED = PYQT5_INSTALLED or PYQT6_INSTALLED or PYSIDE2_INSTALLED or PYSIDE6_INSTALLED

SPECTRAL_CUBE_INSTALLED, requires_spectral_cube = make_skipper('spectral_cube',
                                                               label='spectral-cube')

requires_qt = pytest.mark.skipif(str(not QT_INSTALLED),
                                 reason='An installation of Qt is required')

PYQT_GT_59, _ = make_skipper('PyQt5', version='5.10')

requires_pyqt_gt_59_or_pyside2 = pytest.mark.skipif(str(not PYQT_GT_59 and not PYSIDE2_INSTALLED),
                                                    reason='Requires PyQt > 5.9 or PySide2')


@contextmanager
def make_file(contents, suffix, decompress=False):
    """Context manager to write data to a temporary file,
    and delete on exit

    :param contents: Data to write. string
    :param suffix: File suffix. string
    """
    if decompress:
        contents = zlib.decompress(contents)

    try:
        _, fname = tempfile.mkstemp(suffix=suffix)
        with open(fname, 'wb') as outfile:
            outfile.write(contents)
        yield fname
    finally:
        try:
            os.unlink(fname)
        except WindowsError:  # on Windows the unlink can fail
            pass



# This is a helper function that gets imported into the top-level namespace of glue.

__all__ = ['custom_viewer']


def custom_viewer(name, **kwargs):
    """
    Create a custom interactive data viewer.

    To use this, first create a new variable by calling custom_viewer.
    Then, register one or more viewer functions using decorators.

    :param name: The name of the new viewer
    :type name: str

    Named arguments are used to build widgets and pass data
    to viewer functions. See ``specifying widgets`` below.

    Example::

      v = custom_viewer('My custom viewer', check=False, x='att(x)')

      @v.setup
      def setup_func(axes):
        ''' Setup the plot when the viewer is created '''
        ...

      @v.plot_data
      def plot_data_func(axes, check, style):
          ''' Visualize a full dataset '''
          ...

      @v.plot_subset
      def plot_subset_func(axes, check, style):
          ''' Visualize a subset '''
          ...

      @v.update_settings
      def update_settings_func(check):
          ''' Respond to the user changing a widget setting '''
          ...

      @v.select
      def select(roi, x):
        ''' Filter a dataset based on an roi. Return a boolean array '''
        ...


      @v.make_selector
      def make_selector_func(roi):
          ''' Turn a roi into a subset state '''
          ...


    **Specifying Widgets**

    Keywords passed to ``custom_viewer`` serve two purposes: they
    setup information to be passed into the viewer functions, and
    they create widgets. The type of widget that is created depends
    on the keyword value:

      * ``keyword=False | True`` creates a checkbox. The check state
        is passed as a Boolean into the viewer functions
      * ``keyword=(10, 20, [15])`` creates a slider. The current value
        of the slider is passed as a number to the viewer functions.
        The first two numbers specify the minimum and maximum allowed value,
        while the optional third number specifies the initial value.
      * ``keyword=['a', 'b', 'c']`` creates a dropdown menu. The current
        selection is passed as a string to the viewer functions.
      * ``keyword={'a':1, 'b':2}`` behaves similarly to the lists above,
        but uses the keys as dropdown labels and values as the setting
        passed to viewer functions.
      * ``keyword='att(foo)'`` doesn't create any widget, but passes
        in the attribute named ``foo`` to the viewer functions.
      * ``keyword='att'`` creates a dropdown to let the user select
        one of the attributes from the data.

    **Viewer Functions**

    Custom viewers can implement any of the following functions:

     * ``setup_func`` is called once, when the viewer is created.
     * ``plot_data`` is called to update the visualization of a
       full dataset.
     * ``plot_subset`` is used to visualize data subsets.
     * ``update_settings`` is called whenever a user modifies
       a widget setting.
     * ``select`` specifies how user-drawn regions on the viewer
       are used to filter data. It has access to an :class:`~glue.core.roi.Roi`
       input, and returns a Boolean array testing whether each element
       in a dataset is part of a subset.
     * ``make_selector`` is an alternative to ``select``. Instead of returning
        an array, ``make_selector`` returns a :class:`~glue.core.subset.SubsetState`
    """
    # For now we only support Qt, but this function should be extended to work
    # with non-Qt front-ends in future.
    from glue_qt.viewers.custom import CustomViewer
    return CustomViewer.create_new_subclass(name, **kwargs)



import pytest

import numpy as np
from numpy.testing import assert_equal

from glue.core import Data, DataCollection
from glue.core.link_helpers import LinkSame, ComponentLink
from glue.core.fixed_resolution_buffer import compute_fixed_resolution_buffer

from .test_base_cartesian_data import ExampleBaseData


ARRAY = np.arange(3024).reshape((6, 7, 8, 9)).astype(float)


class TestFixedResolutionBuffer():

    def setup_method(self, method):

        self.data_collection = DataCollection()

        # The reference dataset. Shape is (6, 7, 8, 9).
        self.data1 = Data(x=ARRAY)
        self.data_collection.append(self.data1)

        # A dataset with the same shape but not linked. Shape is (6, 7, 8, 9).
        self.data2 = Data(x=ARRAY)
        self.data_collection.append(self.data2)

        # A dataset with the same number of dimensions but in a different
        # order, linked to the first. Shape is (9, 7, 6, 8).
        self.data3 = Data(x=np.moveaxis(ARRAY, (3, 1, 0, 2), (0, 1, 2, 3)))
        self.data_collection.append(self.data3)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data3.pixel_component_ids[2]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[1],
                                               self.data3.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data3.pixel_component_ids[3]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[3],
                                               self.data3.pixel_component_ids[0]))

        # A dataset with fewer dimensions, linked to the first one. Shape is
        # (8, 7, 6)
        self.data4 = Data(x=ARRAY[:, :, :, 0].transpose())
        self.data_collection.append(self.data4)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data4.pixel_component_ids[2]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[1],
                                               self.data4.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data4.pixel_component_ids[0]))

        # A dataset with even fewer dimensions, linked to the first one. Shape
        # is (8, 6)
        self.data5 = Data(x=ARRAY[:, 0, :, 0].transpose())
        self.data_collection.append(self.data5)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data5.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data5.pixel_component_ids[0]))

        # A dataset that is not on the same pixel grid and requires reprojection
        # self.data6 = Data()
        # self.data6.coords = SimpleCoordinates()
        # self.array_nonaligned = np.arange(60).reshape((5, 3, 4))
        # self.data6['x'] = np.array(self.array_nonaligned)
        # self.data_collection.append(self.data6)
        # self.data_collection.add_link(LinkSame(self.data1.world_component_ids[0],
        #                                        self.data6.world_component_ids[1]))
        # self.data_collection.add_link(LinkSame(self.data1.world_component_ids[1],
        #                                        self.data6.world_component_ids[2]))
        # self.data_collection.add_link(LinkSame(self.data1.world_component_ids[2],
        #                                        self.data6.world_component_ids[0]))

    # Start off with the cases where the data is the target data. Enumerate
    # the different cases for the bounds and the expected result.

    DATA_IS_TARGET_CASES = [

        # Bounds are full extent of data
        ([(0, 5, 6), (0, 6, 7), (0, 7, 8), (0, 8, 9)],
            ARRAY),

        # Bounds are inside data
        ([(2, 3, 2), (3, 3, 1), (0, 7, 8), (0, 7, 8)],
            ARRAY[2:4, 3:4, :, :8]),

        # Bounds are outside data along some dimensions
        ([(-5, 9, 15), (3, 5, 3), (0, 9, 10), (5, 6, 2)],
            np.pad(ARRAY[:, 3:6, :, 5:7], [(5, 4), (0, 0), (0, 2), (0, 0)],
                   mode='constant', constant_values=-np.inf)),

        # No overlap
        ([(2, 3, 2), (3, 3, 1), (-5, -4, 2), (0, 7, 8)],
            -np.inf * np.ones((2, 1, 2, 8)))

    ]

    @pytest.mark.parametrize(('bounds', 'expected'), DATA_IS_TARGET_CASES)
    def test_data_is_target_full_bounds(self, bounds, expected):

        buffer = self.data1.compute_fixed_resolution_buffer(target_data=self.data1, bounds=bounds,
                                                            target_cid=self.data1.id['x'])
        assert_equal(buffer, expected)

        buffer = self.data3.compute_fixed_resolution_buffer(target_data=self.data1, bounds=bounds,
                                                            target_cid=self.data3.id['x'])
        assert_equal(buffer, expected)


def test_base_cartesian_data():

    # Make sure that things work properly when using a BaseCartesianData
    # subclass that isn't necessarily Data.

    data1 = ExampleBaseData(cid_label='x')
    data2 = ExampleBaseData(cid_label='y')

    dc = DataCollection([data1, data2])

    def add_one(x):
        return x + 1

    def sub_one(x):
        return x - 1

    link1 = ComponentLink([data1.pixel_component_ids[0]], data2.pixel_component_ids[0])
    link2 = ComponentLink([data1.pixel_component_ids[1]], data2.pixel_component_ids[1], using=add_one, inverse=sub_one)
    link3 = ComponentLink([data1.pixel_component_ids[2]], data2.pixel_component_ids[2], using=sub_one, inverse=add_one)

    dc.add_link(link1)
    dc.add_link(link2)
    dc.add_link(link3)

    assert_equal(compute_fixed_resolution_buffer(data1,
                                                 target_data=data1,
                                                 bounds=[(-1, 1, 3), (0, 3, 4), 1],
                                                 target_cid=data1.main_components[0]),
                 np.array([[-np.inf, -np.inf, -np.inf, -np.inf],
                           [1, 4, 7, 10],
                           [-np.inf, -np.inf, -np.inf, -np.inf]]))

    assert_equal(compute_fixed_resolution_buffer(data2,
                                                 target_data=data2,
                                                 bounds=[(-1, 1, 3), (0, 3, 4), 1],
                                                 target_cid=data2.main_components[0]),
                 np.array([[-np.inf, -np.inf, -np.inf, -np.inf],
                           [1, 4, 7, 10],
                           [-np.inf, -np.inf, -np.inf, -np.inf]]))

    assert_equal(compute_fixed_resolution_buffer(data1,
                                                 target_data=data2,
                                                 bounds=[(-1, 1, 3), (0, 3, 4), 1],
                                                 target_cid=data1.main_components[0]),
                 np.array([[-np.inf, -np.inf, -np.inf, -np.inf],
                           [-np.inf, 2, 5, 8],
                           [-np.inf, -np.inf, -np.inf, -np.inf]]))



import numpy as np
from numpy.testing import assert_allclose

from astropy.utils import NumpyRNGContext

from glue.tests.visual.helpers import visual_test
from glue.viewers.histogram.viewer import SimpleHistogramViewer
from glue.core.application_base import Application
from glue.core.data import Data


@visual_test
def test_simple_viewer():

    # Make sure the simple viewer can be instantiated

    with NumpyRNGContext(12345):

        data1 = Data(x=np.random.normal(1, 2, 1000), label='data1')
        data2 = Data(y=np.random.uniform(-1, 5, 1000), label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleHistogramViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.id['x'] > 2)

    return viewer.figure


def test_remove_data_collection():

    # Regression test for a bug that caused an IncompatibleAttribute
    # error when updating the number of bins in a histogram after
    # removing a dataset from the DataCollection (this was due to
    # a caching issue)

    data1 = Data(x=[1, 2, 3], label='data1')
    data2 = Data(y=[1, 2, 3], label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleHistogramViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    viewer.state.hist_n_bin = 30

    app.data_collection.remove(data1)

    viewer.state.hist_n_bin = 20


def test_incompatible_datasets():

    # Regression test for a bug that caused an IncompatibleAttribute
    # error when changing the dataset used in the histogram viewer to one that
    # is not linked to the first dataset.

    data1 = Data(x=[1, 2, 3], label='data1')
    data2 = Data(y=[1, 2, 3], label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleHistogramViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    viewer.state.x_att = data1.id['x']

    viewer.state.hist_n_bin = 30

    viewer.state.x_att = data2.id['y']

    viewer.state.hist_n_bin = 20


def test_reset_limits():

    data1 = Data(x=np.arange(1000), label='data')

    app = Application()
    app.data_collection.append(data1)

    viewer = app.new_data_viewer(SimpleHistogramViewer)
    viewer.add_data(data1)

    viewer.state.reset_limits()

    assert_allclose(viewer.state.x_min, 0)
    assert_allclose(viewer.state.x_max, 999)

    viewer.state.x_limits_percentile = 90

    viewer.state.reset_limits()

    assert_allclose(viewer.state.x_min, 49.95)
    assert_allclose(viewer.state.x_max, 949.05)

    assert_allclose(viewer.state.hist_x_min, 49.95)
    assert_allclose(viewer.state.hist_x_max, 949.05)

    viewer.state.update_bins_on_reset_limits = False

    viewer.state.x_limits_percentile = 80

    viewer.state.reset_limits()

    assert_allclose(viewer.state.x_min, 99.9)
    assert_allclose(viewer.state.x_max, 899.1)

    assert_allclose(viewer.state.hist_x_min, 49.95)
    assert_allclose(viewer.state.hist_x_max, 949.05)



import queue
import string
from functools import partial, reduce

from echo.callback_container import CallbackContainer


__all__ = ['DeferredMethod', 'nonpartial', 'lookup_class', 'as_variable_name',
           'as_list', 'file_format', 'CallbackMixin', 'PropertySetMixin',
           'Pointer', 'common_prefix', 'queue_to_list', 'format_choices']


class DeferredMethod(object):
    """
    This class stubs out a method, and provides a callable interface that logs
    its calls. These can later be actually executed on the original
    (non-stubbed) method by calling executed_deferred_calls
    """

    def __init__(self, method):
        self.method = method
        self.calls = []  # avoid hashability issues with dict/set

    @property
    def original_method(self):
        return self.method

    def __call__(self, instance, *a, **k):
        if instance not in (c[0] for c in self.calls):
            self.calls.append((instance, a, k))

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return partial(self.__call__, instance)

    def execute_deferred_calls(self):
        for instance, args, kwargs in self.calls:
            self.method(instance, *args, **kwargs)


def nonpartial(func, *args, **kwargs):
    """
    Like functools.partial, this returns a function which, when called, calls
    ``func(*args, **kwargs)``.

    Unlike functools.partial, extra arguments passed to the returned function
    are *not* passed to the input function.

    This is used when connecting slots to ``QAction.triggered`` signals, which
    appear to have different signatures, which seem to add and extra argument
    in PyQt but not PySide
    """
    def result(*a, **k):
        return func(*args, **kwargs)

    return result


def lookup_class(ref):
    """
    Look up an object via its module string (e.g., 'glue.core.data.Data')

    Parameters
    ----------
    ref : str
        The module string
    """

    if ref.startswith('__builtin__'):
        ref = '.'.join(['builtins'] + ref.split('.')[1:])

    mod = ref.rsplit('.', 1)[0]

    try:
        result = __import__(mod)
    except ImportError:
        raise ValueError("Module '{0}' not found".format(mod))

    try:
        for attr in ref.split('.')[1:]:
            result = getattr(result, attr)
        return result
    except AttributeError:
        raise ValueError("Object '{0}' not found".format(ref))


def as_variable_name(x):
    """
    Convert a string to a legal python variable name

    Parameters
    ----------
    x : str
        A string to (possibly) rename

    Returns
    -------
    variable_name : str
        A legal Python variable name
    """
    allowed = string.ascii_letters + string.digits + '_'
    result = [letter if letter in allowed else '_' for letter in x or 'x']
    if result[0] in string.digits:
        result.insert(0, '_')
    return ''.join(result)


def as_list(x):
    if isinstance(x, list):
        return x
    return [x]


def file_format(filename):
    if filename.find('.') == -1:
        return ''
    if filename.lower().endswith('.gz'):
        result = filename.lower().rsplit('.', 2)[1]
    else:
        result = filename.lower().rsplit('.', 1)[1]
    return result


class CallbackMixin(object):

    """
    A mixin that provides a utility for attaching callback
    functions to methods
    """

    def __init__(self):
        self._callbacks = CallbackContainer()

    def add_callback(self, function):
        self._callbacks.append(function)

    def remove_callback(self, function):
        self._callbacks.remove(function)

    def notify(self, *args, **kwargs):
        for func in self._callbacks:
            func(*args, **kwargs)


class PropertySetMixin(object):

    """An object that provides a set of properties that
    are meant to encapsulate state information

    This class exposes a properties attribute, which is a dict
    of all properties. Similarly, assigning to the properties dict
    will update the individual properties
    """
    _property_set = []

    @property
    def properties(self):
        """ A dict mapping property names to values """
        return dict((p, getattr(self, p)) for p in self._property_set)

    @properties.setter
    def properties(self, value):
        """ Update the properties with a new dict.

        Keys in the new dict must be valid property names defined in
        the _property_set class level attribute"""
        invalid = set(value.keys()) - set(self._property_set)
        if invalid:
            raise ValueError("Invalid property values: %s" % invalid)

        for k in self._property_set:
            if k not in value:
                continue
            setattr(self, k, value[k])


def common_prefix(strings, exclude_punctuation=True):
    """
    Given a list of strings, find the longest prefix common to all of them
    """
    if len(strings) > 0:
        for i in range(len(strings[0]), 0, -1):
            if exclude_punctuation and strings[0][i - 1] in string.punctuation:
                continue
            for st in strings[1:]:
                if st[:i] != strings[0][:i]:
                    break
            else:
                return strings[0][:i]
    return ''


class Pointer(object):

    def __init__(self, key):
        self.key = key

    def __get__(self, instance, type=None):
        val = instance
        for k in self.key.split('.'):
            val = getattr(val, k, None)
        return val

    def __set__(self, instance, value):
        v = self.key.split('.')
        attr = reduce(getattr, [instance] + v[:-1])
        setattr(attr, v[-1], value)


def queue_to_list(q):
    """
    Get all the values in a :class:`queue.Queue` object and return a list.
    """
    l = []
    while True:
        try:
            l.append(q.get_nowait())
        except queue.Empty:
            return l


def format_choices(options, index=False):
    """
    Return a string with an error message formatted as:

    * option 1
    * option 2

    This can be preprended to existing error messages.
    """
    updated_options = []
    for option in options:
        if isinstance(option, str):
            updated_options.append("'{0}'".format(option))
        elif isinstance(option, type):
            updated_options.append(str(option.__module__) + '.' + option.__name__)
        else:
            updated_options.append(option)
    if index:
        return "\n\n" + '\n'.join(['* {0} or {1}'.format(idx, option) for idx, option in enumerate(updated_options)])
    else:
        return "\n\n" + '\n'.join(['* {0}'.format(option) for option in updated_options])



from glue.tests.visual.helpers import visual_test
from glue.viewers.profile.viewer import SimpleProfileViewer
from glue.core.application_base import Application
from glue.core.data import Data


@visual_test
def test_simple_viewer():

    # Make sure the simple viewer can be instantiated

    data1 = Data(x=[1, 2, 3], label='data1')
    data2 = Data(y=[1, 2, 3], label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleProfileViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.pixel_component_ids[0] > 0.8)

    viewer.state.layers[2].linewidth = 5

    return viewer.figure




"""Declare any extra link functions like this"""
# @link_function(info='translates A to B', output_labels=['b'])
# def a_to_b(a):
#    return a * 3


"""Data factories take a filename as input and return a Data object"""
# @data_factory('JPEG Image')
# def jpeg_reader(file_name):
#    ...
#    return data



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

import operator

import pytest
import numpy as np
from unittest.mock import MagicMock

from astropy.wcs import WCS
from shapely.geometry import MultiPolygon, Polygon, Point, LineString

from glue import core
from glue.tests.helpers import requires_astropy

from ..coordinates import Coordinates
from ..component import (Component, DerivedComponent, CoordinateComponent,
                         CategoricalComponent, ExtendedComponent)
from ..component_id import ComponentID
from ..data import Data
from ..parse import ParsedCommand, ParsedComponentLink
from ..data_collection import DataCollection

VIEWS = (np.s_[:], np.s_[1], np.s_[::-1], np.s_[0, :])


class TestComponent(object):

    def setup_method(self, method):
        self.data = MagicMock()
        self.data.shape = [1, 2]
        self.component = Component(self.data)

    def test_data(self):
        assert self.component.data is self.data

    def test_shape(self):
        assert self.component.shape is self.data.shape

    def test_ndim(self):
        assert self.component.ndim is len(self.data.shape)


class TestComponentID(object):

    def setup_method(self, method):
        self.cid = ComponentID('test')

    def test_label(self):
        assert self.cid.label == 'test'

    def test_str(self):
        """ str should return """
        str(self.cid)

    def test_repr(self):
        """ str should return """
        repr(self.cid)


class TestDerivedComponent(object):

    def setup_method(self, method):
        data = MagicMock()
        link = MagicMock()
        self.cid = DerivedComponent(data, link)
        self.link = link
        self.data = data

    def test_data(self):
        """ data property should wrap to links compute method """
        self.cid.data
        self.link.compute.assert_called_once_with(self.data)

    def test_link(self):
        assert self.cid.link == self.link


class TestCategoricalComponent(object):

    def setup_method(self, method):
        self.list_data = ['a', 'a', 'b', 'b']
        self.array_data = np.array(self.list_data)

    def test_autodetection(self):
        assert Component.autotyped(self.array_data).categorical
        assert Component.autotyped(self.list_data).categorical

        x = np.array([True, False, True, False])
        assert not Component.autotyped(x).categorical

        x = np.array([1, 2, 3, 4])
        assert not Component.autotyped(x).categorical

        x = np.array(['1', '2', '3', '4'])
        assert not Component.autotyped(x).categorical

        d = Data(x=['a', 'b', 'c'])
        assert d.get_component('x').categorical

    def test_basic_properties(self):
        data = ['a', 'b', 'c', 'b', 'b', 'c', 'a', 'c', 'd']
        cat_comp = CategoricalComponent(data)
        np.testing.assert_equal(cat_comp.data, data)
        np.testing.assert_equal(cat_comp.labels, data)
        np.testing.assert_equal(cat_comp.codes, [0, 1, 2, 1, 1, 2, 0, 2, 3])
        np.testing.assert_equal(cat_comp.categories, ['a', 'b', 'c', 'd'])

    def test_accepts_numpy(self):
        cat_comp = CategoricalComponent(self.array_data)
        assert cat_comp.data.shape == (4,)

    def test_accepts_pandas(self):
        # Regression test for a bug that caused read-only
        # pandas DataFrame columns to cause issues
        from pandas import DataFrame
        df = DataFrame()
        df['a'] = ['a', 'b', 'c']
        CategoricalComponent(df['a'].fillna(''))

    def test_accepts_list(self):
        """Should accept a list and convert to numpy!"""
        cat_comp = CategoricalComponent(self.list_data)
        np.testing.assert_equal(cat_comp.data, self.array_data)

    def test_multi_nans(self):
        cat_comp = CategoricalComponent(['', '', 'a', 'b', 'c', 'zanthia'])
        np.testing.assert_equal(cat_comp.codes,
                                np.array([0, 0, 1, 2, 3, 4]))
        np.testing.assert_equal(cat_comp.categories,
                                np.asarray(['', 'a', 'b', 'c', 'zanthia']))

    def test_calculate_grouping(self):
        cat_comp = CategoricalComponent(self.array_data)
        np.testing.assert_equal(cat_comp.categories, np.asarray(['a', 'b']))
        np.testing.assert_equal(cat_comp.codes, np.array([0, 0, 1, 1]))
        assert cat_comp.codes.dtype == float

    def test_accepts_provided_grouping(self):
        ncategories = ['b', 'c']
        cat_data = list('aaabbbcccddd')
        cat_comp = CategoricalComponent(cat_data, categories=ncategories)

        assert cat_comp.categories == ncategories
        assert np.all(np.isnan(cat_comp.codes[:3]))
        assert np.all(cat_comp.codes[3:6] == 0)
        assert np.all(cat_comp.codes[6:9] == 1)
        assert np.all(np.isnan(cat_comp.codes[9:]))

    def test_uniform_jitter(self):
        cat_comp = CategoricalComponent(self.array_data)
        second_comp = CategoricalComponent(self.array_data)
        cat_comp.jitter(method='uniform')
        assert np.all(cat_comp.codes != second_comp.codes), "Didn't jitter data!"

    def test_unjitter_data(self):
        cat_comp = CategoricalComponent(self.array_data)
        second_comp = CategoricalComponent(self.array_data)

        cat_comp.jitter(method='uniform')
        delta = np.abs(cat_comp.codes - second_comp.codes).sum()
        assert delta > 0

        cat_comp.jitter(method=None)
        np.testing.assert_equal(cat_comp.codes,
                                second_comp.codes,
                                "Didn't un-jitter data!")

    def test_jitter_on_init(self):
        cat_comp = CategoricalComponent(self.array_data, jitter='uniform')
        second_comp = CategoricalComponent(self.array_data)
        assert np.all(cat_comp.codes != second_comp.codes)

    def test_object_dtype(self):
        d = np.array([1, 3, 3, 1, 'a', 'b', 'a'], dtype=object)
        c = CategoricalComponent(d)

        np.testing.assert_array_equal(c.categories,
                                      np.array([1, 3, 'a', 'b'], dtype=object))
        np.testing.assert_array_equal(c.codes, [0, 1, 1, 0, 2, 3, 2])

    def test_valueerror_on_bad_jitter(self):

        with pytest.raises(ValueError):
            cat_comp = CategoricalComponent(self.array_data)
            cat_comp.jitter(method='this will never be a jitter method')


def test_nd_categorical_component():
    data = [['a', 'b'], ['c', 'b']]
    cat_comp = CategoricalComponent(data)
    np.testing.assert_equal(cat_comp.data, data)
    np.testing.assert_equal(cat_comp.labels, data)
    np.testing.assert_equal(cat_comp.codes, [[0, 1], [2, 1]])
    np.testing.assert_equal(cat_comp.categories, ['a', 'b', 'c'])


def test_nd_categorical_component_autotype():
    x = np.array([['M', 'F'], ['F', 'M']])
    assert Component.autotyped(x).categorical

    d = Data(x=[['male', 'female', 'male'], ['female', 'female', 'male']])
    assert d.get_component('x').categorical


class TestCoordinateComponent(object):

    def setup_method(self, method):

        class TestCoords(Coordinates):

            def __init__(self):
                super().__init__(pixel_n_dim=3, world_n_dim=3)

            def pixel_to_world_values(self, *args):
                return [a * (i + 1) for i, a in enumerate(args)]

            def world_to_pixel_values(self, *args):
                return [a / (i + 1) for i, a in enumerate(args)]

            @property
            def axis_correlation_matrix(self):
                return np.identity(3).astype(bool)

        data = core.Data()
        data.add_component(Component(np.zeros((3, 3, 3))), 'test')
        data.coords = TestCoords()
        self.data = data
        self.px = CoordinateComponent(data, 2)
        self.py = CoordinateComponent(data, 1)
        self.pz = CoordinateComponent(data, 0)
        self.wx = CoordinateComponent(data, 2, world=True)
        self.wy = CoordinateComponent(data, 1, world=True)
        self.wz = CoordinateComponent(data, 0, world=True)

    def test_data(self):
        z, y, x = np.mgrid[0:3, 0:3, 0:3]
        np.testing.assert_array_equal(self.px.data, x)
        np.testing.assert_array_equal(self.py.data, y)
        np.testing.assert_array_equal(self.pz.data, z)
        np.testing.assert_array_equal(self.wx.data, x * 1)
        np.testing.assert_array_equal(self.wy.data, y * 2)
        np.testing.assert_array_equal(self.wz.data, z * 3)

    @pytest.mark.parametrize(('view'), VIEWS)
    def test_view(self, view):
        z, y, x = np.mgrid[0:3, 0:3, 0:3]

        np.testing.assert_array_equal(self.px[view], x[view])
        np.testing.assert_array_equal(self.py[view], y[view])
        np.testing.assert_array_equal(self.pz[view], z[view])
        np.testing.assert_array_equal(self.wx[view], x[view] * 1)
        np.testing.assert_array_equal(self.wy[view], y[view] * 2)
        np.testing.assert_array_equal(self.wz[view], z[view] * 3)


def check_binary(result, left, right, op):
    assert isinstance(result, core.subset.InequalitySubsetState)
    assert result.left is left
    assert result.right is right
    assert result.operator is op


def check_link(result, left, right):
    assert isinstance(result, core.component_link.ComponentLink)
    if isinstance(left, ComponentID):
        assert left in result.get_from_ids()
    if isinstance(right, ComponentID):
        assert right in result.get_from_ids()


# componentID overload
COMPARE_OPS = (operator.gt, operator.ge, operator.lt, operator.le)
NUMBER_OPS = (operator.add, operator.mul, operator.truediv, operator.sub)


@pytest.mark.parametrize(('op'), COMPARE_OPS)
def test_inequality_scalar(op):
    cid = ComponentID('test')
    result = op(cid, 3)
    check_binary(result, cid, 3, op)


@pytest.mark.parametrize(('op'), COMPARE_OPS)
def test_inequality_id(op):
    cid = ComponentID('test')
    cid2 = ComponentID('test2')
    result = op(cid, cid2)
    check_binary(result, cid, cid2, op)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_scalar(op):
    cid = ComponentID('test')
    result = op(cid, 3)
    check_link(result, cid, 3)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_scalar_right(op):
    cid = ComponentID('test')
    result = op(3, cid)
    check_link(result, 3, cid)


@pytest.mark.parametrize(('op'), NUMBER_OPS)
def test_arithmetic_cid(op):
    cid = ComponentID('test')
    cid2 = ComponentID('test2')
    result = op(cid, cid2)
    check_link(result, cid, cid2)


def test_pow_scalar():
    cid = ComponentID('test')
    result = cid ** 3
    check_link(result, cid, 3)


@pytest.mark.parametrize(('view'), VIEWS)
def test_view(view):
    comp = Component(np.array([[1, 2, 3], [2, 3, 4]]))
    np.testing.assert_array_equal(comp[view], comp.data[view])


@pytest.mark.parametrize(('view'), VIEWS)
def test_view_derived(view):
    comp = Component(np.array([[1, 2, 3], [2, 3, 4]]))
    d = core.Data()
    cid = d.add_component(comp, 'primary')
    cid2 = ComponentID("derived")
    link = core.ComponentLink([cid], cid2, using=lambda x: x * 3)
    dc = DerivedComponent(d, link)

    np.testing.assert_array_equal(dc[view], comp.data[view] * 3)


@requires_astropy
def test_units():

    # Make sure that units get converted to strings. At the moment if these
    # are set to Astropy units for example, things can go wrong for example
    # when writing out the datasets. Once we settle on a units framework, we
    # can then use that instead of converting units to strings.

    from astropy import units as u

    comp = Component([1, 2, 3], units='m')
    assert comp.units == 'm'
    assert isinstance(comp.units, str)

    comp = Component([1, 2, 3], units=u.m)
    assert comp.units == 'm'
    assert isinstance(comp.units, str)


def test_scalar_derived():
    # Regression test for a bug that caused expressions without components to
    # fail. Note that the data collection is needed here as the bug occurs in
    # the link manager set up by the data collection.
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('3.5', {})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [3.5, 3.5, 3.5])


def test_rename_cid_used_in_derived():
    # Unit test to make sure expressions still work when renaming components
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('1 + {a}', {'a': data.id['a']})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [4, 5, 2])
    data.id['a'].label = 'x'
    np.testing.assert_equal(data['b'], [4, 5, 2])


@pytest.mark.xfail
def test_update_cid_used_in_derived():
    # Unit test to make sure expressions still work when updating the component
    # id for components
    data = Data(a=[3, 4, 1])
    dc = DataCollection([data])  # noqa
    pc = ParsedCommand('1 + {a}', {'a': data.id['a']})
    link = ParsedComponentLink(ComponentID('b'), pc)
    data.add_component_link(link)
    np.testing.assert_equal(data['b'], [4, 5, 2])
    data.update_id(data.id['a'], ComponentID('x'))
    np.testing.assert_equal(data['b'], [4, 5, 2])


def test_coordinate_component_1d_coord():

    # Regression test for a bug that caused incorrect world coordinate values
    # for 1D coordinates.

    wcs = WCS(naxis=1)
    wcs.wcs.ctype = ['FREQ']
    wcs.wcs.crpix = [1]
    wcs.wcs.crval = [1]
    wcs.wcs.cdelt = [1]

    data = Data(flux=np.random.random(5), coords=wcs, label='data')
    np.testing.assert_equal(data['Frequency'], [1, 2, 3, 4, 5])


class TestExtendedComponent(object):

    def setup_method(self):

        self.cen_x_id = ComponentID('x')
        self.cen_y_id = ComponentID('y')

        poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
        poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
        poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
        poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

        polygons = MultiPolygon([poly_3, poly_4])
        self.polys = np.array([poly_1, poly_2, polygons])
        self.poly2d = ExtendedComponent(self.polys, center_comp_ids=[self.cen_x_id, self.cen_y_id])

        circle_1 = Point(1.0, 0.0).buffer(1)
        circle_2 = Point(2.0, 3.0).buffer(2)
        circles = np.array([circle_1, circle_2])
        self.circles = ExtendedComponent(circles, center_comp_ids=[self.cen_x_id, self.cen_y_id])

        ranges = np.array([LineString([(0, 0), (1, 0)]), LineString([(0, 0), (4, 0)])])
        self.ranges = ExtendedComponent(ranges, center_comp_ids=[self.cen_x_id])

    def test_basic_proprties(self):
        assert self.poly2d.ndim == 1
        assert isinstance(self.poly2d, ExtendedComponent)
        assert self.poly2d.shape == (3,)
        assert self.poly2d.x == self.cen_x_id
        assert self.poly2d.y == self.cen_y_id

        assert self.ranges.ndim == 1
        assert isinstance(self.ranges, ExtendedComponent)
        assert self.ranges.shape == (2,)
        assert self.ranges.x == self.cen_x_id
        assert self.ranges.y is None

    def test_incorrect_inputs(self):
        with pytest.raises(TypeError, match='Input data for a ExtendedComponent should be a list of shapely.Geometry objects'):
            bad_data = np.array([1, 2, 3])
            bad_data_comp = ExtendedComponent(bad_data, center_comp_ids=[self.cen_x_id, self.cen_y_id])

        with pytest.raises(ValueError, match='ExtendedComponent must be initialized with one or two ComponentIDs'):
            no_center_ids_comp = ExtendedComponent(self.polys, center_comp_ids=[])

        with pytest.raises(ValueError, match='ExtendedComponent must be initialized with one or two ComponentIDs'):
            no_center_ids_comp = ExtendedComponent(self.polys, center_comp_ids=[self.cen_x_id, self.cen_y_id, self.cen_x_id])



from glue.config import stretches
from glue.viewers.matplotlib.state import (
    DeferredDrawDictCallbackProperty as DDDCProperty,
    DeferredDrawSelectionCallbackProperty as DDSCProperty,
)

__all__ = ["StretchStateMixin"]


class StretchStateMixin:
    stretch = DDSCProperty(
        docstring="The stretch used to render the layer, "
        "which should be one of ``linear``, "
        "``sqrt``, ``log``, or ``arcsinh``"
    )
    stretch_parameters = DDDCProperty(
        docstring="Keyword arguments to pass to the stretch"
    )

    _stretch_set_up = False

    def setup_stretch_callback(self):
        type(self).stretch.set_choices(self, list(stretches.members))
        type(self).stretch.set_display_func(self, stretches.display_func)
        self._reset_stretch()
        self.add_callback("stretch", self._reset_stretch)
        self.add_callback("stretch_parameters", self._sync_stretch_parameters)
        self._stretch_set_up = True

    @property
    def stretch_object(self):
        if not self._stretch_set_up:
            raise Exception("setup_stretch_callback has not been called")
        return self._stretch_object

    def _sync_stretch_parameters(self, *args):
        for key, value in self.stretch_parameters.items():
            if hasattr(self._stretch_object, key):
                setattr(self._stretch_object, key, value)
            else:
                raise ValueError(
                    f"Stretch object {self._stretch_object.__class__.__name__} has no attribute {key}"
                )

    def _reset_stretch(self, *args):
        self._stretch_object = stretches.members[self.stretch]()
        self.stretch_parameters.clear()



# Tests for the translation infrastrcture between glue Data/Subset/SubsetState
# objects and other data containers registered with glue.

import pytest
import numpy as np
from numpy.testing import assert_equal

from ..component_id import ComponentID
from ..data import Data
from ..data_collection import DataCollection
from ..registry import Registry

from glue.config import data_translator, subset_state_translator
from ..subset import InequalitySubsetState


class FakeDataObject(object):
    array = None
    name = None


class AnotherFakeDataObject(object):
    pass


class CustomSelectionObject(object):
    serialized = None


def setup_module(self):

    @data_translator(FakeDataObject)
    class FakeDataObjectHandler:

        def to_data(self, obj):
            data = Data()
            data[obj.name] = obj.array
            return data

        def to_object(self, data_or_subset):
            cid = data_or_subset.main_components[0]
            obj = FakeDataObject()
            obj.array = data_or_subset[cid]
            obj.name = cid.label
            return obj

    @subset_state_translator('my_subset_translator')
    class FakeSubsetDefinitionTranslator:
        """
        A simple subset translator that only knows how to handle
        InequalitySubsetState and translates it to a custom serialization.
        We include a keyword argument for the translation to make sure it
        gets passed through.
        """

        def to_object(self, subset):

            subset_state = subset.subset_state

            if isinstance(subset_state, InequalitySubsetState):

                if isinstance(subset_state.left, ComponentID):
                    left = '{' + subset_state.left.label + '}'
                else:
                    left = subset_state.left

                if isinstance(subset_state.right, ComponentID):
                    right = '{' + subset_state.right.label + '}'
                else:
                    right = subset_state.right

                c = CustomSelectionObject()
                c.serialized = '{0} {1} {2}'.format(left, subset_state.operator.__name__, right)

                return c

            else:
                raise TypeError("my_subset_translator could not translate "
                                "subset state of type {0}".format(subset_state.__class__.__name__))


def teardown_module(self):
    data_translator.remove(FakeDataObject)
    subset_state_translator.remove('my_subset_translator')


class TestTranslationData:

    def teardown_method(self, method):
        Registry().clear()

    def test_get_object_basic(self):

        data = Data(spam=np.arange(10))

        obj = data.get_object(cls=FakeDataObject)
        assert isinstance(obj, FakeDataObject)
        assert_equal(obj.array, np.arange(10))
        assert_equal(obj.name, 'spam')

    def test_get_object_invalid(self):

        data = Data(spam=np.arange(10))

        with pytest.raises(TypeError) as exc:
            data.get_object(cls=AnotherFakeDataObject)
        assert exc.value.args[0] == 'Could not find a class to translate objects of type Data to AnotherFakeDataObject'

    def test_get_object_explicit_class(self):

        data = Data(x=[1, 2, 3], label='myobj')

        with pytest.raises(ValueError) as exc:
            data.get_object()
        assert 'Specify the object class to use with cls' in exc.value.args[0]
        assert 'glue.core.tests.test_data_translation.FakeDataObject' in exc.value.args[0]

        obj = data.get_object(cls=FakeDataObject)
        assert isinstance(obj, FakeDataObject)
        assert_equal(obj.array, [1, 2, 3])
        assert_equal(obj.name, 'x')

    def test_get_subset_object(self):

        data = Data(spam=np.arange(10))
        data.add_subset(data.id['spam'] > 4.5, label='subset 1')

        # Check that the following three are equivalent
        for subset_id in [None, 0, 'subset 1']:

            subset = data.get_subset_object(subset_id=subset_id, cls=FakeDataObject)

            assert isinstance(subset, FakeDataObject)
            assert_equal(subset.array, np.arange(5, 10))
            assert subset.name == 'spam'

    def test_get_subset_object_explicit_class(self):

        data = Data(spam=np.arange(10))
        data.add_subset(data.id['spam'] > 4.5, label='subset 1')

        with pytest.raises(ValueError) as exc:
            data.get_subset_object()
        assert 'Specify the object class to use with cls' in exc.value.args[0]
        assert 'glue.core.tests.test_data_translation.FakeDataObject' in exc.value.args[0]

        subset = data.get_subset_object(subset_id=0, cls=FakeDataObject)

        assert isinstance(subset, FakeDataObject)
        assert_equal(subset.array, np.arange(5, 10))
        assert subset.name == 'spam'

    def test_get_subset_object_invalid(self):

        data = Data(x=np.arange(10), label='myobj')

        with pytest.raises(ValueError) as exc:
            data.get_subset_object(subset_id='subset 2', cls=FakeDataObject)
        assert exc.value.args[0] == "Dataset does not contain any subsets"

        data.add_subset(data.id['x'] > 4.5, label='subset 1')
        data.add_subset(data.id['x'] > 3.5, label='subset 1')

        with pytest.raises(ValueError) as exc:
            data.get_subset_object(subset_id='subset 2', cls=FakeDataObject)
        assert exc.value.args[0] == "No subset found with the label 'subset 2'"

        with pytest.raises(ValueError) as exc:
            data.get_subset_object(cls=FakeDataObject)
        assert exc.value.args[0] == "Several subsets are present, specify which one to retrieve with subset_id= - valid options are:\n\n* 0 or 'subset 1'\n* 1 or 'subset 1_01'"

        # FIXME: currently we disambiguate subset names, but would maybe be
        # better to just raise an error. In any case the test below never
        # works because of the auto-disambiguation.
        # with pytest.raises(ValueError) as exc:
        #     data.get_subset_object(subset_id='subset 1', cls=FakeDataObject)
        # assert exc.value.args[0] == "Several subsets were found with the label 'subset 1', use a numerical index instead"

        subset = data.get_subset_object(subset_id=0, cls=FakeDataObject)

        assert isinstance(subset, FakeDataObject)
        assert_equal(subset.array, np.arange(5, 10))
        assert subset.name == 'x'

    def test_get_selection_basic(self):

        data = Data(x=[1, 2, 3], label='basic')
        data.add_subset(data.id['x'] > 1, label='subset 1')

        # Check that the following three are equivalent
        for subset_id in [None, 0, 'subset 1']:

            result = data.get_selection_definition(subset_id=subset_id,
                                                   format='my_subset_translator')

            assert isinstance(result, CustomSelectionObject)
            assert result.serialized == '{x} gt 1'

    def test_get_selection_invalid(self):

        data = Data(x=[1, 2, 3], label='basic')

        with pytest.raises(ValueError) as exc:
            data.get_selection_definition(subset_id=0)
        assert exc.value.args[0] == "Dataset does not contain any subsets"

        data.add_subset((data.id['x'] > 1) & (data.id['x'] < 3), label='subset 1')

        with pytest.raises(TypeError) as exc:
            data.get_selection_definition(subset_id=0, format='my_subset_translator')
        assert exc.value.args[0] == 'my_subset_translator could not translate subset state of type AndState'

        with pytest.raises(ValueError) as exc:
            data.get_selection_definition(subset_id=0)
        assert exc.value.args[0] == ("Subset state handler format not set - should be one "
                                     "of:\n\n* 'my_subset_translator'")

        with pytest.raises(ValueError) as exc:
            data.get_selection_definition(subset_id=0, format='invalid_translator')
        assert exc.value.args[0] == ("Invalid subset state handler format 'invalid_translator' "
                                     "- should be one of:\n\n* 'my_subset_translator'")

        data.add_subset((data.id['x'] > 1) & (data.id['x'] < 3), label='subset 1')
        data.add_subset((data.id['x'] > 1) & (data.id['x'] < 3), label='subset 3')

        with pytest.raises(ValueError) as exc:
            data.get_selection_definition(subset_id='subset 2', format='invalid_translator')
        assert exc.value.args[0] == "No subset found with the label 'subset 2'"

        with pytest.raises(ValueError) as exc:
            data.get_selection_definition(format='invalid_translator')
        assert exc.value.args[0] == ("Several subsets are present, specify which "
                                     "one to retrieve with subset_id= - valid options "
                                     "are:\n\n* 0 or 'subset 1'\n* 1 or 'subset 1_01'\n* 2 or 'subset 3'")

        # with pytest.raises(ValueError) as exc:
        #     data.get_selection_definition(subset_id='subset 1', format='invalid_translator')
        # assert exc.value.args[0] == "Several subsets were found with the label 'subset 1', use a numerical index instead"


class TestTranslationDataCollection:

    # The main purpose of these tests is to make sure that glue remembers the
    # original class from objects that are added to the data collection.

    def setup_method(self, method):
        self.dc = DataCollection()

    def test_set_and_get_object_basic(self):

        obj = FakeDataObject()
        obj.array = np.arange(10)
        obj.name = 'spam'

        assert len(self.dc) == 0
        self.dc['myobj'] = obj
        assert len(self.dc) == 1
        assert self.dc['myobj'] is self.dc[0]

        data = self.dc['myobj']
        assert isinstance(data, Data)
        assert_equal(data['spam'], np.arange(10))

        # Make sure preferred class is used
        obj_out = self.dc['myobj'].get_object()
        assert isinstance(obj_out, FakeDataObject)
        assert obj_out is not obj
        assert_equal(obj_out.array, np.arange(10))
        assert_equal(obj_out.name, 'spam')

    def test_set_object_invalid(self):

        obj = AnotherFakeDataObject()

        with pytest.raises(TypeError) as exc:
            self.dc['myobj'] = obj
        assert exc.value.args[0] == 'Could not find a class to translate objects of type AnotherFakeDataObject to Data'

    def test_get_subset_object(self):

        obj = FakeDataObject()
        obj.array = np.arange(10)
        obj.name = 'spam'

        self.dc['myobj'] = obj
        self.dc.new_subset_group(subset_state=self.dc['myobj'].id['spam'] > 4.5,
                                 label='subset 1')

        subset = self.dc['myobj'].get_subset_object(subset_id=0)

        # Make sure preferred class is used
        assert isinstance(subset, FakeDataObject)
        assert_equal(subset.array, np.arange(5, 10))
        assert subset.name == 'spam'

    def test_set_duplicate(self):

        # Make sure data gets replaced when re-using a data collection key

        assert len(self.dc) == 0
        self.dc['myobj'] = Data(x=[1, 2, 3])
        assert len(self.dc) == 1
        assert_equal(self.dc[0]['x'], [1, 2, 3])
        self.dc['myobj'] = Data(x=[4, 5, 6])
        assert len(self.dc) == 1
        assert_equal(self.dc[0]['x'], [4, 5, 6])

        # And make sure if there are multiple datasets with the key
        # pre-existing then we get rid of them all

        self.dc.append(Data(x=[7, 8, 9], label='myobj'))
        assert len(self.dc) == 2
        self.dc['myobj'] = Data(x=[1, 2, 3])
        assert len(self.dc) == 1
        assert_equal(self.dc[0]['x'], [1, 2, 3])



import numpy as np

from glue.core.units import UnitConverter
from glue.core.subset import roi_to_subset_state

from glue.viewers.matplotlib.viewer import SimpleMatplotlibViewer
from glue.viewers.profile.state import ProfileViewerState
from glue.viewers.profile.layer_artist import ProfileLayerArtist

__all__ = ['MatplotlibProfileMixin', 'SimpleProfileViewer']


class MatplotlibProfileMixin(object):

    def setup_callbacks(self):
        self.state.add_callback('x_att', self._update_axes)
        self.state.add_callback('normalize', self._update_axes)
        self.state.add_callback('y_display_unit', self._update_axes)

    def _update_axes(self, *args):

        if self.state.x_att is not None:
            self.state.x_axislabel = self.state.x_att.label

        if self.state.normalize:
            self.state.y_axislabel = 'Normalized data values'
        else:
            if self.state.y_display_unit:
                self.state.y_axislabel = f'Data values [{self.state.y_display_unit}]'
            else:
                self.state.y_axislabel = 'Data values'

        self.axes.figure.canvas.draw_idle()

    def apply_roi(self, roi, override_mode=None):

        # Force redraw to get rid of ROI. We do this because applying the
        # subset state below might end up not having an effect on the viewer,
        # for example there may not be any layers, or the active subset may not
        # be one of the layers. So we just explicitly redraw here to make sure
        # a redraw will happen after this method is called.
        self.redraw()

        if len(self.layers) == 0:
            return

        # Apply inverse unit conversion, converting from display to native units
        converter = UnitConverter()
        cmin, cmax = converter.to_native(self.state.reference_data,
                                         self.state.x_att, np.array([roi.min, roi.max]),
                                         self.state.x_display_unit)

        # Sometimes unit conversions can cause the min/max to be swapped
        if cmin > cmax:
            cmin, cmax = cmax, cmin

        roi.min = cmin
        roi.max = cmax

        subset_state = roi_to_subset_state(roi, x_att=self.state.x_att)
        self.apply_subset_state(subset_state, override_mode=override_mode)


class SimpleProfileViewer(MatplotlibProfileMixin, SimpleMatplotlibViewer):

    _state_cls = ProfileViewerState
    _data_artist_cls = ProfileLayerArtist
    _subset_artist_cls = ProfileLayerArtist

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        MatplotlibProfileMixin.setup_callbacks(self)



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u

from glue.core import ComponentID, Data, DataCollection
from glue.config import settings, unit_converter

from .. import link_helpers as lh
from ..link_helpers import (LinkTwoWay, MultiLink,
                            LinkSame, LinkAligned,
                            LinkSameWithUnits)


def setup_function(func):
    func.ORIGINAL_UNIT_CONVERTER = settings.UNIT_CONVERTER


def teardown_function(func):
    settings.UNIT_CONVERTER = func.ORIGINAL_UNIT_CONVERTER


def teardown_module():
    unit_converter._members.pop('test-spectral1')


R, D, L, B = (ComponentID('ra'), ComponentID('dec'),
              ComponentID('lon'), ComponentID('lat'))


def forwards(x):
    print('forwards input', x)
    return x * 3


def backwards(x):
    print('backwards input', x)
    return x / 3


def forwards_xy(x, y):
    print('forwards inputs', x, y)
    return x * 3, y * 5


def backwards_xy(x, y):
    print('backwards inputs', x, y)
    return x / 3, y / 5


def check_link(link, from_, to, using=None):
    assert link.get_from_ids() == from_
    assert link.get_to_id() == to
    if using:
        assert link.get_using() == using


def check_using(link, inp, out):
    assert_allclose(link.get_using()(*inp), out)


def test_LinkTwoWay():
    result = list(LinkTwoWay(R, D, forwards, backwards))
    check_link(result[0], [R], D, forwards)
    check_link(result[1], [D], R, backwards)


def test_multilink_forwards():
    result = MultiLink([R, D], [L, B], forwards_xy, labels2=['x', 'y'])
    assert len(result) == 2
    check_link(result[0], [R, D], L)
    check_link(result[1], [R, D], B)
    check_using(result[0], (3, 4), 9)
    check_using(result[1], (3, 4), 20)


def test_multilink_backwards():
    result = MultiLink([R, D], [L, B], backwards=backwards_xy, labels1=['x', 'y'])
    assert len(result) == 2
    check_link(result[0], [L, B], R)
    check_link(result[1], [L, B], D)
    check_using(result[0], (9, 20), 3)
    check_using(result[1], (9, 20), 4)


def test_multilink_forwards_backwards():
    result = MultiLink([R, D], [L, B], forwards_xy, backwards_xy)
    assert len(result) == 4
    check_link(result[0], [R, D], L)
    check_link(result[1], [R, D], B)
    check_link(result[2], [L, B], R)
    check_link(result[3], [L, B], D)
    check_using(result[0], (3, 4), 9)
    check_using(result[1], (3, 4), 20)
    check_using(result[2], (9, 20), 3)
    check_using(result[3], (9, 20), 4)


def test_multilink_nofunc():
    with pytest.raises(TypeError) as exc:
        MultiLink([R, D], [L, B])
    assert exc.value.args[0] == "Must supply either forwards or backwards"


def test_linksame_string():
    """String inputs auto-converted to component IDs"""
    # ComponentLink does type checking to ensure conversion happens
    links = LinkSame('a', 'b')


def test_identity():
    assert lh.identity('3') == '3'


def test_toid():
    assert lh._toid('test').label == 'test'

    cid = ComponentID('test2')
    assert lh._toid(cid) is cid

    with pytest.raises(TypeError) as exc:
        lh._toid(None)


@pytest.mark.parametrize('ndim', [0, 1, 2])
def test_link_aligned(ndim):
    shp = tuple([2] * ndim)
    data1 = Data(test=np.random.random(shp))
    data2 = Data(test=np.random.random(shp))

    links = LinkAligned(data1, data2)
    dc = DataCollection([data1, data2])
    dc.add_link(links)

    for i in range(ndim):
        id0 = data1.pixel_component_ids[i]
        id1 = data2.pixel_component_ids[i]
        assert_allclose(data1[id0], data2[id1])


def test_link_units():

    data1 = Data(weight_kg=[1, 2, 3])
    data2 = Data(weight_g=[2000, 3000, 1000])

    data1.get_component(data1.id['weight_kg']).units = 'kg'
    data2.get_component(data2.id['weight_g']).units = 'g'

    links = LinkSameWithUnits(data1.id['weight_kg'], data2.id['weight_g'])
    dc = DataCollection([data1, data2])
    dc.add_link(links)

    assert_allclose(data1[data1.id['weight_kg']], [1, 2, 3])
    assert_allclose(data2[data1.id['weight_kg']], [2, 3, 1])
    assert_allclose(data1[data2.id['weight_g']], [1000, 2000, 3000])
    assert_allclose(data2[data2.id['weight_g']], [2000, 3000, 1000])


def test_link_units_invalid():

    data1 = Data(weight_kg=[1, 2, 3])
    data2 = Data(distance_m=[2000, 3000, 1000])

    data1.get_component(data1.id['weight_kg']).units = 'kg'
    data2.get_component(data2.id['distance_m']).units = 'm'

    with pytest.raises(u.UnitConversionError, match='not convertible'):
        LinkSameWithUnits(data1.id['weight_kg'], data2.id['distance_m'])


@unit_converter('test-spectral1')
class SpectralUnitConverter:

    def equivalent_units(self, data, cid, units):
        return map(str, u.Unit(units).find_equivalent_units(include_prefix_units=True, equivalencies=u.spectral()))

    def to_unit(self, data, cid, values, original_units, target_units):
        return (values * u.Unit(original_units)).to_value(target_units, equivalencies=u.spectral())


def test_link_units_custom():

    data1 = Data(wav_mm=[1, 2, 3])
    data2 = Data(freq_ghz=[10, 20, 30])

    data1.get_component(data1.id['wav_mm']).units = 'mm'
    data2.get_component(data2.id['freq_ghz']).units = 'GHz'

    with pytest.raises(u.UnitConversionError, match='not convertible'):
        LinkSameWithUnits(data1.id['wav_mm'], data2.id['freq_ghz'])

    settings.UNIT_CONVERTER = 'test-spectral1'

    links = LinkSameWithUnits(data1.id['wav_mm'], data2.id['freq_ghz'])
    dc = DataCollection([data1, data2])
    dc.add_link(links)

    assert_allclose(data1[data1.id['wav_mm']], [1, 2, 3])
    assert_allclose(data2[data1.id['wav_mm']], [29.9792, 14.9896, 9.9930], rtol=1e-3)
    assert_allclose(data1[data2.id['freq_ghz']], [299.7924, 149.8962, 99.9308], rtol=1e-3)
    assert_allclose(data2[data2.id['freq_ghz']], [10, 20, 30])



# Configuration file for the Sphinx documentation builder.

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Glue"
copyright = "2012-2023, Chris Beaumont, Thomas Robitaille, Michelle Borkin"
author = "Chris Beaumont, Thomas Robitaille, Michelle Borkin"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autoclass_content = "both"

nitpick_ignore = [
    ("py:class", "glue.viewers.histogram.layer_artist.HistogramLayerBase"),
    ("py:class", "glue.viewers.scatter.layer_artist.ScatterLayerBase"),
    ("py:class", "glue.viewers.image.layer_artist.ImageLayerBase"),
    ("py:class", "glue.viewers.image.layer_artist.RGBImageLayerBase"),
    ("py:class", "glue.viewers.image.state.BaseImageLayerState"),
    ("py:class", "glue.viewers.common.stretch_state_mixin.StretchStateMixin")
]

viewcode_follow_imported_members = False

numpydoc_show_class_members = False
autosummary_generate = True
automodapi_toctreedirnm = "api"

linkcheck_ignore = [r"https://s3.amazonaws.com"]
linkcheck_retries = 5
linkcheck_timeout = 10


intersphinx_mapping = {
    "python": ("https://docs.python.org/3.7", None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "echo": ("https://echo.readthedocs.io/en/latest/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "shapely": ("https://shapely.readthedocs.io/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_logo = "_static/logo.png"
html_theme_options = {'navigation_with_keys': False}



# Functions used to parse data and links from Python command-line

import numpy as np

from glue.config import cli_parser
from glue.core import BaseData, Data

__all__ = []


@cli_parser(dict)
def _parse_data_dict(data, label):
    result = Data(label=label)
    for label, component in data.items():
        result.add_component(component, label)
    return [result]


@cli_parser(np.recarray)
def _parse_data_recarray(data, label):
    kwargs = dict((n, data[n]) for n in data.dtype.names)
    return [Data(label=label, **kwargs)]


@cli_parser(BaseData)
def _parse_data_glue_data(data, label):
    if isinstance(data, Data):
        data.label = label
    return [data]


@cli_parser(np.ndarray)
def _parse_data_numpy(data, label):
    return [Data(**{label: data, 'label': label})]


@cli_parser(list)
def _parse_data_list(data, label):
    return [Data(**{label: data, 'label': label})]


@cli_parser(str)
def _parse_data_path(path, label):
    from glue.core.data_factories import load_data, as_list

    data = load_data(path)
    for d in as_list(data):
        d.label = label
    return as_list(data)


def parse_data(data, label):

    # First try new data translation layer

    from glue.config import data_translator

    try:
        handler, preferred = data_translator.get_handler_for(data)
    except TypeError:
        pass
    else:
        data = handler.to_data(data)
        data.label = label
        data._preferred_translation = preferred
        return [data]

    # Then try legacy 'cli_parser' infrastructure

    for item in cli_parser:

        data_class = item.data_class
        parser = item.parser
        if isinstance(data, data_class):
            try:
                return parser(data, label)
            except Exception as e:
                raise ValueError("Invalid format for data '%s'\n\n%s" %
                                 (label, e))

    raise TypeError("Invalid data description: %s" % data)


def parse_links(dc, links):
    from glue.core.link_helpers import MultiLink
    from glue.core import ComponentLink

    data = dict((d.label, d) for d in dc)
    result = []

    def find_cid(s):
        dlabel, clabel = s.split('.')
        d = data[dlabel]
        c = d.find_component_id(clabel)
        if c is None:
            raise ValueError("Invalid link (no component named %s)" % s)
        return c

    for link in links:
        f, t = link[0:2]  # from and to component names
        u = u2 = None
        if len(link) >= 3:  # forward translation function
            u = link[2]
        if len(link) == 4:  # reverse translation function
            u2 = link[3]

        # component names -> component IDs
        if isinstance(f, str):
            f = [find_cid(f)]
        else:
            f = [find_cid(item) for item in f]

        if isinstance(t, str):
            t = find_cid(t)
            result.append(ComponentLink(f, t, u))
        else:
            t = [find_cid(item) for item in t]
            result += MultiLink(f, t, u, u2)

    return result



import uuid
import weakref

import numpy as np

import matplotlib.patches as mpatches
import matplotlib.colors as mcolors

from glue.utils import defer_draw

from glue.viewers.image.state import ImageLayerState, ImageSubsetLayerState
from glue.viewers.image.python_export import python_export_image_layer, python_export_image_subset_layer
from glue.viewers.image.pixel_selection_subset_state import PixelSubsetState
from glue.viewers.matplotlib.layer_artist import MatplotlibLayerArtist
from glue.core.exceptions import IncompatibleAttribute
from glue.utils import color2rgb
from glue.utils.matplotlib import ColormapPatchHandler
from glue.core import BaseData, HubListener
from glue.core.message import (ComponentsChangedMessage,
                               ExternallyDerivableComponentsChangedMessage,
                               PixelAlignedDataChangedMessage)
from glue.viewers.image.frb_artist import imshow
from glue.core.fixed_resolution_buffer import ARRAY_CACHE, PIXEL_CACHE


class BaseImageLayerArtist(MatplotlibLayerArtist, HubListener):

    def __init__(self, axes, viewer_state, layer_state=None, layer=None):

        super(BaseImageLayerArtist, self).__init__(axes, viewer_state,
                                                   layer_state=layer_state, layer=layer)

        # Watch for changes in the viewer state which would require the
        # layers to be redrawn
        self._viewer_state.add_global_callback(self._update_image)
        self.state.add_global_callback(self._update_image)

        self.layer.hub.subscribe(self, ComponentsChangedMessage,
                                 handler=self.update,
                                 filter=self._is_data_object)

        self.layer.hub.subscribe(self, ExternallyDerivableComponentsChangedMessage,
                                 handler=self.update,
                                 filter=self._is_data_object)

        self.layer.hub.subscribe(self, PixelAlignedDataChangedMessage,
                                 handler=self.update,
                                 filter=self._is_data_object)

    def _is_data_object(self, message):
        if isinstance(self.layer, BaseData):
            return message.sender is self.layer
        else:
            return message.sender is self.layer.data

    def _update_image(self, force=False, **kwargs):
        raise NotImplementedError()


class ImageLayerArtist(BaseImageLayerArtist):

    _layer_state_cls = ImageLayerState
    _python_exporter = python_export_image_layer

    def __init__(self, axes, viewer_state, layer_state=None, layer=None):

        super(ImageLayerArtist, self).__init__(axes, viewer_state,
                                               layer_state=layer_state, layer=layer)

        # We use a custom object to deal with the compositing of images, and we
        # store it as a private attribute of the axes to make sure it is
        # accessible for all layer artists.
        self.uuid = str(uuid.uuid4())
        self.composite = self.axes._composite
        self.composite.allocate(self.uuid)
        self.composite.set(self.uuid, array=self.get_image_data,
                           shape=self.get_image_shape)
        self.composite_image = self.axes._composite_image

    @property
    def label(self):
        return "%s (%s)" % (self.layer.label, self.state.attribute.label)

    def get_layer_color(self):
        if self._viewer_state.color_mode == 'One color per layer':
            return self.state.color
        else:
            return self.state.cmap

    def get_handle_legend(self):
        if self.enabled and self.state.visible:
            # a fancier patch drawing the colormap
            color = self.get_layer_color()
            label = self.layer.label
            if isinstance(color, mcolors.Colormap):
                handle = mpatches.Patch(color=color(0.5))
                # actual color of the handle is not important
                handler = ColormapPatchHandler(color)
            else:
                handle = mpatches.Patch(color=color, alpha=self.layer.style.alpha)
                handler = None  # default handler is good enough
            return handle, label, handler
        else:
            return None, None, None

    def enable(self):
        if not self.enabled and hasattr(self, 'composite_image'):
            self.composite_image.invalidate_cache()
        super(ImageLayerArtist, self).enable()

    def remove(self):
        super(ImageLayerArtist, self).remove()
        if self.uuid in self.composite:
            self.composite.deallocate(self.uuid)
            self.composite_image.invalidate_cache()
        ARRAY_CACHE.pop(self.state.uuid, None)
        PIXEL_CACHE.pop(self.state.uuid, None)
        self.uuid = None

    def get_image_shape(self):

        if self.uuid is None or self._viewer_state.reference_data is None or self._viewer_state.x_att is None or self._viewer_state.y_att is None:
            return None

        x_axis = self._viewer_state.x_att.axis
        y_axis = self._viewer_state.y_att.axis

        full_shape = self._viewer_state.reference_data.shape

        return full_shape[y_axis], full_shape[x_axis]

    def get_image_data(self, bounds=None):

        if self.uuid is None:
            return None

        try:
            image = self.state.get_sliced_data(bounds=bounds)
        except (IncompatibleAttribute, IndexError):
            # The following includes a call to self.clear()
            self.disable_invalid_attributes(self.state.attribute)
            return None
        else:
            self.enable()

        return image

    def _update_image_data(self):
        self.composite_image.invalidate_cache()
        self.redraw()

    @defer_draw
    def _update_visual_attributes(self):

        if not self.enabled:
            return

        if self._viewer_state.color_mode == 'Colormaps':
            self.composite.mode = 'colormap'
        else:
            self.composite.mode = 'color'

        self.composite.set(self.uuid,
                           clim=(self.state.v_min, self.state.v_max),
                           visible=self.state.visible,
                           zorder=self.state.zorder,
                           color=self.state.color,
                           cmap=self.state.cmap,
                           contrast=self.state.contrast,
                           bias=self.state.bias,
                           alpha=self.state.alpha,
                           stretch=self.state.stretch_object)

        self.composite_image.invalidate_cache()

        self.redraw()

    @defer_draw
    def _update_image(self, force=False, **kwargs):

        if self.uuid is None or self.state.attribute is None or self.state.layer is None:
            return

        # NOTE: we need to evaluate this even if force=True so that the cache
        # of updated properties is up to date after this method has been called.
        changed = self.pop_changed_properties()

        if force or any(prop in changed for prop in ('layer', 'attribute',
                                                     'slices', 'x_att', 'y_att')):
            self._update_image_data()
            force = True  # make sure scaling and visual attributes are updated

        if force or any(prop in changed for prop in ('v_min', 'v_max', 'contrast',
                                                     'bias', 'alpha', 'color_mode',
                                                     'cmap', 'color', 'zorder',
                                                     'visible', 'stretch', 'stretch_parameters')):
            self._update_visual_attributes()

    @defer_draw
    def update(self, *event):
        ARRAY_CACHE.pop(self.state.uuid, None)
        PIXEL_CACHE.pop(self.state.uuid, None)
        self._update_image(force=True)
        self.redraw()


class ImageSubsetArray(object):

    def __init__(self, viewer_state, layer_artist):
        self._viewer_state = weakref.ref(viewer_state)
        self._layer_artist = weakref.ref(layer_artist)
        self._layer_state = weakref.ref(layer_artist.state)

    @property
    def layer_artist(self):
        return self._layer_artist()

    @property
    def layer_state(self):
        return self._layer_state()

    @property
    def viewer_state(self):
        return self._viewer_state()

    @property
    def shape(self):

        x_axis = self.viewer_state.x_att.axis
        y_axis = self.viewer_state.y_att.axis

        full_shape = self.viewer_state.reference_data.shape

        return full_shape[y_axis], full_shape[x_axis]

    def __call__(self, bounds):

        if (self.layer_artist is None or
                self.layer_state is None or
                self.viewer_state is None):
            return None

        # We should compute the mask even if the layer is not visible as we need
        # the layer to show up properly when it is made visible (which doesn't
        # trigger __getitem__). However, if the layer is disabled, then we will
        # call this method when it is enabled, so in this case we just return
        # an empty mask.

        try:
            mask = self.layer_state.get_sliced_data(bounds=bounds)
        except IncompatibleAttribute:
            self.layer_artist.disable_incompatible_subset()
            return None
        else:
            self.layer_artist.enable(redraw=False)

        r, g, b = color2rgb(self.layer_state.color)
        mask = np.dstack((r * mask, g * mask, b * mask, mask * .5))
        mask = (255 * mask).astype(np.uint8)

        return mask

    @property
    def dtype(self):
        return np.uint8

    @property
    def ndim(self):
        return 2

    @property
    def size(self):
        return np.prod(self.shape)


class ImageSubsetLayerArtist(BaseImageLayerArtist):

    _layer_state_cls = ImageSubsetLayerState
    _python_exporter = python_export_image_subset_layer

    def __init__(self, axes, viewer_state, layer_state=None, layer=None):

        super(ImageSubsetLayerArtist, self).__init__(axes, viewer_state,
                                                     layer_state=layer_state, layer=layer)

        self.subset_array = ImageSubsetArray(self._viewer_state, self)

        self.image_artist = imshow(self.axes, self.subset_array,
                                   origin='lower', interpolation='nearest',
                                   vmin=0, vmax=1, aspect='auto')

        self._line_x = self.axes.axvline(0)
        self._line_x.set_visible(False)

        self._line_y = self.axes.axhline(0)
        self._line_y.set_visible(False)

        self.mpl_artists = [self.image_artist, self._line_x, self._line_y]

    @defer_draw
    def _update_data(self):
        if isinstance(self.state.layer.subset_state, PixelSubsetState):
            try:
                x, y = self.state.layer.subset_state.get_xy(self.layer.data,
                                                            self._viewer_state.x_att.axis,
                                                            self._viewer_state.y_att.axis)
            except IncompatibleAttribute:
                self._line_x.set_visible(False)
                self._line_y.set_visible(False)
            else:
                self._line_x.set_data([x, x], [0, 1])
                self._line_x.set_visible(True)
                self._line_y.set_data([0, 1], [y, y])
                self._line_y.set_visible(True)
        else:
            self._line_x.set_visible(False)
            self._line_y.set_visible(False)
        self.image_artist.invalidate_cache()

    @defer_draw
    def _update_visual_attributes(self, redraw=True):

        if not self.enabled:
            return

        for artist in self.mpl_artists:
            if artist is self.image_artist:
                artist.set_visible(self.state.visible)
                artist.set_alpha(self.state.alpha)
            else:
                if self.state.visible:
                    artist.set_visible(isinstance(self.state.layer.subset_state, PixelSubsetState))
                else:
                    artist.set_visible(False)
                artist.set_color(self.state.color)
                artist.set_alpha(self.state.alpha * 0.5)
            artist.set_zorder(self.state.zorder)

        if redraw:
            self.redraw()

    def _update_image(self, force=False, **kwargs):

        if self.state.layer is None:
            return

        # NOTE: we need to evaluate this even if force=True so that the cache
        # of updated properties is up to date after this method has been called.
        changed = self.pop_changed_properties()

        if force or any(prop in changed for prop in ('layer', 'attribute', 'color',
                                                     'x_att', 'y_att', 'slices')):
            self._update_data()
            force = True  # make sure scaling and visual attributes are updated

        if force or any(prop in changed for prop in ('zorder', 'visible', 'alpha')):
            self._update_visual_attributes()

    def remove(self):
        super(ImageSubsetLayerArtist, self).remove()
        self.image_artist.invalidate_cache()
        ARRAY_CACHE.pop(self.state.uuid, None)
        PIXEL_CACHE.pop(self.state.uuid, None)

    def enable(self, redraw=True):
        super(ImageSubsetLayerArtist, self).enable()
        # We need to now ensure that image_artist, which may have been marked
        # as not being visible when the layer was cleared is made visible
        # again.
        if hasattr(self, 'image_artist'):
            self.image_artist.invalidate_cache()
            self._update_visual_attributes(redraw=redraw)

    @defer_draw
    def update(self, *event):
        ARRAY_CACHE.pop(self.state.uuid, None)
        PIXEL_CACHE.pop(self.state.uuid, None)
        self._update_image(force=True)
        self.redraw()



"""
Test that data.update_components() sends a NumericalDataChangedMessage
that conveys which components have been changed.
"""
from glue.core.data import Data
from glue.core.hub import HubListener
from glue.core.data_collection import DataCollection
from glue.core.message import NumericalDataChangedMessage

import numpy as np
from numpy.testing import assert_array_equal


def test_message_carries_components():

    test_data = Data(x=np.array([1, 2, 3, 4, 5]), y=np.array([1, 2, 3, 4, 5]), label='test_data')
    data_collection = DataCollection([test_data])

    class CustomListener(HubListener):

        def __init__(self, hub):
            self.received = 0
            self.components_changed = None
            hub.subscribe(self, NumericalDataChangedMessage,
                            handler=self.receive_message)

        def receive_message(self, message):
            self.received += 1
            try:
                self.components_changed = message.components_changed
            except AttributeError:
                self.components_changed = None

    listener = CustomListener(data_collection.hub)
    assert listener.received == 0
    assert listener.components_changed is None

    cid_to_change = test_data.id['x']
    new_data = [5, 2, 6, 7, 10]
    test_data.update_components({cid_to_change: new_data})

    assert listener.received == 1
    assert cid_to_change in listener.components_changed

    assert_array_equal(test_data['x'], new_data)
